#ifndef GF_MATERIAL_H
#define GF_MATERIAL_H

#include <model/gattrib.h>
#include <util/gstring.h>
#include <util/gsurface.h>
#include <util/gtempl.h>
#include <render/gsolidt.h>

class GFTexture: public GSolidTexture {
    // CONCEPT: Furniture graphics kernel
    //
    // ABSTRACT PURPOSE: Base class of GF solid textures.
    //
    // INVARIANT: 
    // 
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
    // 
  protected:
    GFTexture(void): mScale(1.0), mRatio(0.3) {}
  public:
    virtual const GSurface& getSurface(const GPoint3 &, const GSurface&) const = 0;

#ifdef GPRE_ANSI
    void setScale(GFloat pScale) { mScale = pScale; }
#else
    void setScale(GFloat pScale) const { mScale = pScale; }
#endif
    GFloat getScale(void) const {
	return(mScale);
    }
#ifdef GPRE_ANSI
    void setRatio(GFloat pRatio) { mRatio = pRatio; }
#else
    void setRatio(GFloat pRatio) const { mRatio = pRatio; }
#endif
    GFloat getRatio(void) const {
	return(mRatio);
    }
  private:
#ifdef GPRE_ANSI
    GFloat mScale;
    GFloat mRatio;
#else
    mutable GFloat mScale;
    mutable GFloat mRatio;
#endif
};

class GFZeroTexture: public GFTexture {
    // CONCEPT: Furniture graphics kernel
    //
    // ABSTRACT PURPOSE: A wood-like texture w/ concentric ringes in y direction.
    // 
    // INVARIANT: 
    // 
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
    // 
  public:
    GFZeroTexture(void) {}
    virtual const GSurface& getSurface(const GPoint3 &, const GSurface& pSurface) const {
	return(pSurface);
    } 
};

class GFWoodTexture: public GFTexture {
    // CONCEPT: Furniture graphics kernel
    //
    // ABSTRACT PURPOSE: A wood-like texture w/ concentric ringes in y direction.
    // 
    // INVARIANT: 
    // 
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
    // 
  public:
    GFWoodTexture(void) {}
    virtual const GSurface& getSurface(const GPoint3 &, const GSurface&) const;
};

class GFSurface: public GGraphicalData, public GSurface {
    // CONCEPT: Furniture graphics kernel
    //
    // ABSTRACT PURPOSE: A utility type for storing surface properties.
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
    // 
  private:
    GFSurface(const GString&); 
  public:
    const GString& getName(void) const { return(mName); }
    void setScale(GFloat pScale) { 
	mScale = pScale; 
    }
    void setRatio(GFloat pRatio) { 
	mRatio = pRatio; 
    }
    void setTexture(const GFTexture& pTexture) {
	mTexture = &pTexture;
    }
    const GFTexture& getTexture(void) const {
#ifdef GPRE_ANSI
	((GFTexture*)mTexture)->setScale(mScale);
	((GFTexture*)mTexture)->setRatio(mRatio);
#else
	mTexture->setScale(mScale);
	mTexture->setRatio(mRatio);
#endif
	return(*mTexture);
    }
  public:
    static const GFSurface& getByName(const GString&);
    static const GFSurface& getDefault(void) {
	return(sDefault);
    }
  private:
    // list of loaded surfaces:
    static GDLList<GFSurface> sList;
    // default surface:
    static const GFSurface sDefault;
  private:
    const GString mName;
    // a scaling factor for texture:
    GFloat mScale;
    // a color ratio factor for texture:
    GFloat mRatio;

    const GFTexture* mTexture;

    static GFZeroTexture sZero;
    static GFWoodTexture sWood;
};

class GFMaterial {
    // CONCEPT: Furniture graphics kernel
    //
    // ABSTRACT PURPOSE: Specific material attribute
    // handling symbolic names. Materials are loaded
    // from a data base.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: 
    //
    // CONSTANTS: 
  public:
    GFMaterial(void);
    GFMaterial(const GString&);
    GFMaterial(const GFMaterial&);
    GFMaterial(const GFSurface&);
    
    // Set and get the surface:
    void setMaterial(const GFSurface& pSurface) { 
	mSurface = &pSurface; 
    }
    const GFSurface& getMaterial(void) const { 
	return(*mSurface); 
    }
    const GSurface& getSurface(void) const { 
	return(*mSurface); 
    }

    // Set from symbolic name, return true if successful:
    bool setMaterial(const GString&); 

    // Get symbolic material name:
    GString getName(void) const { 
	GString s(10);
	s = "{";
	s += mSurface->getName();
	s += "}";
	return(s);
    }
    bool operator==(const GFMaterial&) const;
  private:
    const GFSurface* mSurface;
};

ostream& operator<<(ostream&, const GFMaterial&);

inline bool GFMaterial::operator==(const GFMaterial& pMaterial) const {
    return(mSurface == pMaterial.mSurface ? true : false);
}

class GFMaterialAttribute: public GAttribute {
    // CONCEPT: Furniture graphics kernel
    //
    // ABSTRACT PURPOSE: Encapsulation of the 
    // material attribute.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GFMaterialAttribute(void);
    GFMaterialAttribute(const GFMaterial&);

    bool operator==(const GFMaterialAttribute&) const;

    void setMaterial(const GFMaterial&);
    const GFMaterial& getMaterial(void) const;

    void reference(const GFMaterialAttribute&);
  protected:
    virtual void materialChanged(void) {}
  private:
    // the default instance of this attribute type:
    static const GFMaterialAttribute attribute;
    // a dynamically associated data instance:
    GFMaterial* mMaterial;
};

inline bool GFMaterialAttribute::operator==(const GFMaterialAttribute& pMaterial) const {
    return(getMaterial() == pMaterial.getMaterial() ? true : false);
}

#endif
