#ifndef GF_PRIMITIVE_H
#define GF_PRIMITIVE_H

#include <ggeneric.h>
#include <gprimitv.h>
#include <util/gident.h>
#include <util/gtcl.h>
#include <model/gmodel.h>
#include <model/gresol.h>

#include "gfmaterl.h"
#include "gfrender.h"

extern bool gStoreCompletely;
// If the boolean is true print the attributes even if they are not physically existent.

class GFPrimitive;

extern const GFPrimitive* gfNoFather;

typedef enum {
    // CONCEPT: Furniture graphics kernel
    //
    // ABSTRACT PURPOSE: Enumeration of selectability
    // hierarchical interaction.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
    GF_NOT_SELECTABLE,
    // The object cannot be selected.
    GF_SELECTABLE,
    // The object itself may be selected.
    GF_SELECT_FATHER
    // Try to select the father or nothing.	
} GFSelectionMode;

typedef enum {
    // CONCEPT: Furniture graphics kernel
    //
    // ABSTRACT PURPOSE: Enumeration of 
    // a mode for writing class data.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
    GF_PRIMITIVE,
    // The object is a primitive object
    // and is not built from other ones.
    GF_UNCHANGED,
    // The object hierarchy root by this object
    // was not changed since creation.
    GF_CHANGED
    // The object itself or a child was changed.
} GFWriteMode;

class GFPrimitive: 
public GPrimitive, 
public GIdentification,
#ifdef GWITH_TCL
public GTclInterface,
#endif
public GModelingAttribute, 
public GResolutionAttribute, 
public GFMaterialAttribute, 
public GShadable, 
public GRayTraceable 
{
    // CONCEPT: Furniture graphics kernel
    //
    // ABSTRACT PURPOSE: The basic primitive of the furniture
    // graphics kernel.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: gfNoFather 
  public:
    GFPrimitive(const char* NAME, GFPrimitive* = (GFPrimitive*)gfNoFather);

    virtual void addChild(GFPrimitive* pChild) {
	mChildren.append(pChild);
    }
    
    // type information:
    virtual bool isA(const GString&) const = 0;
    virtual const char* getDescription(void) const = 0;
    virtual const char* getClass(void) const = 0;
    
    // "positional modeling":
    // The position is the point with the lowest x,
    // lowest y and lowest z coordinate of the bounding
    // box in model coordinates.
    // set position relatively:
    void setPosition(const GPoint3&);
    
    // Translation in world coordinates:
    void translate(const GVector3&);
  public:
    void save(ostream&);
    // METHOD: Save ASCII description of object to output stream.
  private:
    virtual void saveConstr(ostream&) = 0;
    // METHOD: Save ASCII representation of the constructor arguments.
    void saveModeling(ostream&);
    // METHOD: Save ASCII representation of physically existent attributes.
    void saveClasses(ostream&);
    // METHOD: Save an "include <class>" for all occuring GFP/GFT classes
    // of this object and its children.
    void beAnElement(void);
    // METHOD: The object is an element of another object.
    void setElement(GFPrimitive*);
    // METHOD: Insert passed primitive into element list and
    // increase element pointer.
  protected:
    virtual void saveChildren(ostream&);
    // METHOD: Save children, too.
    const GString& saveName(bool father = false);
  protected:
    void setChanged(void);
    // Mark recursively the object changed.
    void setFatherChanged(void);
    // Mark recursively the father changed.
    void setUnchanged(void) {
	mWriteMode = GF_UNCHANGED;
    }
    // Mark the object unchanged.
    bool wasChanged(void) const {
	return(mWriteMode == GF_CHANGED);
    }
  public:
#ifdef GWITH_TCL
    const GString& add(const GPoint3& POSITION, const char* CLASS, const char* ARGUMENTS);
    // Add an element object of specified CLASS at specified POSITION by evaluating
    // the passed constructor ARGUMENTS.
    // Returns name of created widget relatively.
#endif
    
    // Remove a primitive from both children
    // and element list:
    void remove(GFPrimitive* p) {
	setChanged();
	mElements.remove(p);
	mChildren.remove(p);
	// mChildren's remove will delete the child, too.
    }

    const GBounds3& getWorldBounds(void) const {
	static GBounds3 sBounds;
	sBounds.setMax(mWCBounds.getMax());
	sBounds.setMin(mWCBounds.getMin());
	return(sBounds);
    }

    GFPrimitive* getFather(void) {
	return(mFather);
    }

    bool isElement(void) const {
	return(mIsElement);
    }

    GDLList<GFPrimitive>& getElements(void) {
	return(mElements);
    }

    GDLList<GFPrimitive>& getChildren(void) {
	return(mChildren);
    }
  public:
    // Shading interface:
    virtual void shade(void);
    // Ray-tracing interface:
    virtual bool intersect(const GRay3&, GInterSectionList&);
    virtual GVector3 getNormal(const GPoint3&);
    virtual const GSurface& getSurface(const GPoint3&);
    virtual GFloat getTransmission(void) const {
	return(getMaterial().getSurface().getTransparency());
    }

    virtual GFloat getNearest(const GRay3& pRay, GShadable*& rShad) {
	GFPrimitive* tPrim = (GFPrimitive*)rShad;
	GFloat tDist = getNearest(pRay, tPrim);
	rShad = (GShadable*)tPrim;
	return(tDist);
    }
    virtual GFloat getNearest(const GRay3& pRay, GRayTraceable*& rRay) {
	GFPrimitive* tPrim = (GFPrimitive*)rRay;
	GFloat tDist = getNearest(pRay, tPrim);
	rRay = (GRayTraceable*)tPrim;
	return(tDist);
    }
    virtual GFloat getNearest(const GRay3&, GFPrimitive*&);

    // compute and return bounding box in WC:
    const GBounds3& computeBounds(void);

    // set/get pickability:
    void setSelectionMode(GFSelectionMode pMode) {
	mMode = pMode;
    }
    GFSelectionMode getSelectionMode(void) const {
	return(mMode);
    }
  protected:
    const GSurface& getMappedSurface(const GPoint3&, const GSurface&);

    virtual void materialChanged(void);
    virtual void resolutionChanged(void);
    virtual void modelingChanged(void);
    // The three methods are transferring changes down.
    // The have to be called in overloaded methods of subclasses.

    GPoint3 getPosition(void) const;
    // METHOD: Inquire the local position.
    GFloat getRotation(void) const;
    // METHOD: Inquire the local y-axis rotation.
  protected:
    virtual GFRenderable* getRenderable(void) {
	return(gfNoRenderable);
    } 
  private:
    void add(GFPrimitive*);
    // METHOD: Add a primitive as element.
    // an internally-used utility method:
    void boundingBox(GVector3&, GFloat, GFloat, GFloat);
    // Get pointer to primitive with specified name:
  public:
    static GFPrimitive* getByName(const char*);
#ifdef GWITH_TCL
  protected:
    static int classCMD(ClientData, Tcl_Interp*, int, char*[]);
    virtual bool objectCMD(char *[]);
#endif
  private:
    GFPrimitive* mFather;
    // all components of the composite:
    GDLList<GFPrimitive> mChildren;
    // list of elements:
    // Elements is a sub-set of the children.
    GDLList<GFPrimitive> mElements;
    // Elements is a sub-set of the children.
    void checkPosition(void);
    int mElCnt;
    GFSelectionMode mMode;
    GRayTraceableBounds mWCBounds; 
    GFWriteMode mWriteMode;
    bool mIsElement;
};

template <class T>
class GFGetNameIterator: public GIterator<T> {
    // CONCEPT: Furniture graphics kernel
    //
    // ABSTRACT PURPOSE: Get names of primitives
    // into a string list.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GFGetNameIterator(void): mList(20) {}

    void start(void) { mList = "{"; }
    void stop(void) { mList += '}'; }

    virtual bool exec(T& t) {
	mList += t.getIdentification();
	mList += ' ';
	return(false);
    }
    const char* get(void) { return(mList); }
  private:
    GString mList;
};

inline void GFPrimitive::setChanged(void) {
    if (mWriteMode == GF_UNCHANGED) {
	mWriteMode = GF_CHANGED;
	if (getFather() != gfNoFather) getFather()->setChanged();
    }
}

inline void GFPrimitive::setFatherChanged(void) {
    if (getFather() != gfNoFather) getFather()->setChanged();
}

template <class T>
class GFSavePrimitiveIterator: public GIterator<T> {
    // CONCEPT: Furniture graphics kernel
    //
    // ABSTRACT PURPOSE: Save primitives to 
    // specified stream.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GFSavePrimitiveIterator(ostream pStream): mStream(pStream) {}
    virtual bool exec(T& t) {
	t.save(mStream);
	return(false);
    }
  private:
    ostream mStream;
};

#endif

