#ifndef GF_RENDER_H
#define GF_RENDER_H

#include <gdata.h>
#include <render/grender.h>
#include <render/gshade.h>
#include <render/graytrac.h>

#include "gfbasic.h"

enum GFRenderEnum {
    // CONCEPT: Furniture graphics kernel
    //
    // ABSTRACT PURPOSE: There is a limited number of basic renderables
    // for this graphics kernel. We\'ll need this following information 
    // for the heterogeneous cache table to perform a type-safe down casting.
    // Only storable renderables need to define this.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
    GF_NO_RENDERABLE,

    GF_CYLINDER,
    GF_FRAME,
    GF_OFF_POLYGON,
    GF_SPHERE,
    GF_SWEEP_X_POLYGON,
    GF_SWEEP_Y_POLYGON,
    GF_SWEEP_Z_POLYGON
}; 

class GFPrimitive;

class GFRenderable: public GShadable, public GRayTraceable, public GGraphicalData {
    // CONCEPT: Furniture graphics kernel
    //
    // ABSTRACT PURPOSE: Root class of the basic
    // primitives of the furniture kernel.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: Multiple references are
    // supported using hash indexes to get a shared represen-
    // tation from data base.
    // CONSTANTS: 
  public:
    typedef GDLList<GFRenderable> list;
    typedef GCacheTableEntry<GFRenderable> entry;
    typedef GDLList<entry> table;
    void setModelingPrimitive(GFPrimitive& pModelPrim) {
	mModelPrim = &pModelPrim;
    }
    GFPrimitive& getModelPrimitive(void) const {
	assert(mModelPrim);
	return(*mModelPrim);
    }
    // Unreference a multiple referenced renderable:
    static void unreference(GFRenderable*);

    // Inquire run-time identification for storable renderables:
    virtual GFRenderEnum getID(void) const { return(GF_NO_RENDERABLE); }

    // Inquire cache value for storable renderables:
    virtual GCacheValue getCacheValue(void) const { return(0); }
  public:
    // rendering interface:
    virtual GFloat getNearest(const GRay3& pRay, GShadable*& rShad) {
	GFRenderable* tRend = (GFRenderable*)rShad;
	GFloat tDist = getNearest(pRay, tRend);
	rShad = (GShadable*)tRend;
	return(tDist);
    }
    virtual GFloat getNearest(const GRay3& pRay, GRayTraceable*& rRay) {
	GFRenderable* tRend = (GFRenderable*)rRay;
	GFloat tDist = getNearest(pRay, tRend);
	rRay = (GRayTraceable*)tRend;
	return(tDist);
    }
    virtual GFloat getNearest(const GRay3&, GFRenderable*&);
  protected:
    // utilitys for WC/MC and vice versa conversion:
    GRay3 ray2mc(const GRay3& WC_RAY);
    GFloat distance2wc(const GRay3& MC_RAY, const GRay3& WC_RAY, GFloat MC_DISTANCE);
    GVector3 normal2wc(const GPoint3& WC_POINT, const GVector3& MC_NORMAL);
  protected:
    static void insertInCache(GFRenderable*);
    static list* getCacheList(GCacheValue);
  private:
    GFPrimitive* mModelPrim;
    static table sTable;
};

extern GFRenderable* gfNoRenderable;

#endif
