
//
// Notes: An object may be present many times in the BSP-hierarchy, so
//        we can not delete the objects directly. For this popurse, a
//        primary list is created at contruction's time. The Dtor
//        uses this list for deallocating the primitives. A side
//        effect is, that we can use this list for BVH or whatever...
//


#include <stdlib.h>
#include "render/graytrac.h"

#include "gt_defs.h"
#include "gt_abox.h"
#include "gt_bbox.h"
#include "gt_stack.h"
#include "gt_bsp.h"

#ifndef SAFE_DIV
#define SAFE_DIV(a, b) ((b)) ? (a)/(b) : MAXFLT
#endif

int __compareMinMax(const void *pA, const void *pB )   {
    GFloat *tA  = (GFloat *) pA;
    GFloat *tB  = (GFloat *) pB;
    if ( *tA == *tB ) 
      return( 0 ) ;
    else   
      return( *tA < *tB ? -1 : 1 ) ;
}

// @FUNCT@: @__subdivide@
// @DESCR@: recursive divide a BSP-Node 
// @NOTES@: boys, this routine is really slow ! But dont bother about it
// because it's pre-processing time. The idea, how to make a balanced
// BSP-Tree, is simple, but very hard to implement without bounding boxes.
static void __subdivide(GT_BSP_Node*  rNode,
			int           pDepth,        
			int           pMaxDepth,     
			int           pMaxListLength )
{
    int i, j, k;
    GT_BSP_Node*  tCur;
    GFloat*       tMins[3];
    GFloat*       tMaxs[3];
    int           tNum;
    int           tAxis;
   
    rNode->mChild[1] = rNode->mChild[0] = 0;
   
    GTObjectList *tListChildren = rNode->getListOfChildren();
    tNum = tListChildren->count();
   
    if ((tNum  > pMaxListLength) && (pDepth < pMaxDepth))  {
      
	// initialize 3 arrays for later sorting.. 
	for(i = 0; i < 3; i++) {
	    tMins[i] = new GFloat[tNum];
	    tMaxs[i] = new GFloat[tNum];
	}
      
	i = 0;
	for(GTObjectItr _Itr(*tListChildren);
	    _Itr.more(); 
	    _Itr.next())
	  {
	      GTABox  tBox(_Itr.cur()->getABox());
	 
	      tMins[X_AXIS][i] = tBox.min().getX();
	      tMins[Y_AXIS][i] = tBox.min().getY();
	      tMins[Z_AXIS][i] = tBox.min().getZ();
	      tMaxs[X_AXIS][i] = tBox.max().getX();
	      tMaxs[Y_AXIS][i] = tBox.max().getY();
	      tMaxs[Z_AXIS][i] = tBox.max().getZ();
	      i++;
	  }
      
	// Sorting.. 
	for(i = 0; i < 3 ; i++) {
	    qsort( tMins[i], tNum, sizeof(GFloat), __compareMinMax) ;
	    qsort( tMaxs[i], tNum, sizeof(GFloat), __compareMinMax) ;
	}
      
	// generate 2 children w/ randomly div-axis! 
	for (i = 0; i < 2; i++)  {
	    tCur = rNode->mChild[i] = new GT_BSP_Node;
	 
	    tCur->min(rNode->min());
	    tCur->max(rNode->max());
	    tCur->mAxis = rand() % 3;
	}
	
	// testing for optimal subdivision 
	bool tFOUND = FALSE;
	for (j = 0; j < tNum && !tFOUND; j++) {
		
	    // for each axis.. 
	    for (tAxis = rNode->mAxis, k = 0; k < 3; 
		 tAxis = (tAxis+1)%3, k++) {
	       
		if (tMins[tAxis][j] >= tMaxs[tAxis][tNum - j - 1]) {
		    GFloat tDiv = 0.5 * (tMins[tAxis][j] + 
					 tMaxs[tAxis][tNum -j - 1]);
		    rNode->mChild[0]->max()[tAxis] = 
		      rNode->mChild[1]->min()[tAxis] = tDiv;
		    
		    rNode->mAxis = tAxis;
		    tFOUND = TRUE;
		    break;
		}
	    }
	}
	 
	// stop subdividing if it doesnt help anymore! 
	if (tFOUND)  {
	    for (i = 0; i < 2; i++)  {
		tCur = rNode->mChild[i];
		for(GTObjectItr _Itr(*tListChildren);
		    _Itr.more(); _Itr.next()) 
		  {
		      GTObject *tObj = _Itr.cur();
		      
		      // simple test for overlaping 
		      // Note: There is no need for expanding the 
		      // parent BSP-bound!
		      if( tObj->isOverlap(*tCur))
			tCur->getListOfChildren()->append(tObj);
		  }
	    
		__subdivide(tCur, pDepth+1, pMaxDepth, pMaxListLength);
	    }
	}
	else {
	    KILL_MEM(rNode->mChild[0]);
	    KILL_MEM(rNode->mChild[1]);
	}
	
	for( i = 0; i < 3; i++) {
	    delete [] tMins[i];
	    delete [] tMaxs[i];
	}
    }
}


static void __buildBVH(GT_BSP_Node *rNode) {
   
    if(!(rNode->mChild[0])) {
	rNode->buildHierarchy();
    } else {
	__buildBVH(rNode->mChild[0]);
	__buildBVH(rNode->mChild[1]);
    }
}


// DTor
GT_BSP_Node::~GT_BSP_Node(void) {
    KILL_MEM(mChild[0]);
    KILL_MEM(mChild[1]);
}


//
// @GT_BSP_Node::classify@
// 
// @DESCR@: find out, in wich child-node the rays origin lies.
// Modify: dist  -> distance from rays origin to the div-plane
//         pNear -> the node, in wich the rays origin lies
//         pFar  -> the other node.
// @NOTES@:
//  There was a bug in the BSP-Code in GGemsIII ! If the rays
//  origin lies in the division plane, we must test the rays
//  direction to classify the right NEAR- and FAR child.
//  
//  The actual BSP-code works in object-space. Modification for
//  ray-space is neccessary by ray-classification-algorithm. 
//
void GT_BSP_Node::classify(const GRay3&  pRay,
			   GFloat&       rDist,
			   GT_BSP_Node*& rNear,
			   GT_BSP_Node*& rFar)
 {
     
     GFloat tOriginToDivPlane = mChild[0]->mMax[mAxis] - (pRay.getPoint())[mAxis];
     GFloat tDivDirection     = (pRay.getVector())[mAxis];
     
     if ( tOriginToDivPlane > 0 ){
	 rNear = mChild[0];
	 rFar  = mChild[1];
     } else if(tOriginToDivPlane < 0){
	 rFar  = mChild[0];
	 rNear = mChild[1];
     } 
     // This bug was really hard to find...
     else {
	 if ( tDivDirection < 0) {
	     rNear = mChild[0];
	     rFar  = mChild[1];
	 } else {
	     rFar  = mChild[0];
	     rNear = mChild[1];
	 }
     }
     rDist =  SAFE_DIV(tOriginToDivPlane, tDivDirection);
 }




 // define a stack for recursion-elimination
class BSP_StkNode  {
 public:
    GT_BSP_Node*  mNode;
    GFloat        mMin, mMax;
   
    BSP_StkNode(GT_BSP_Node* pNode = 0, GFloat pMin = 0, GFloat pMax = 0) {
	mNode = pNode;
	mMin = pMin;
	mMax = pMax;
    }
};
 
//
// @FUNCT@: @GT_BSP_Tree::intersect@
// 
// @DESCR@: intersect the ray w/ the BSP-Hierarchy.
// 
// @NOTES@: This routine is based on the "linear time voxel traversing"
// algorithm from Jim Arvo. Have a look on RTNews2 to see how it works!
// 
bool GT_BSP_Tree::intersect(GRay3&              rRay, 
			    GInterSectionList&  rInter)
{ 
    // define a static stack for traversing 
    static  Stack<BSP_StkNode>  sTraversStack;   
    GT_BSP_Node     *tCur, *tNear, *tFar;
    GFloat          tDist, tTmin, tTmax;
    
    bool tHIT = FALSE;
   
    if (!mRoot->findMinMax(rRay, tTmin, tTmax))
      return FALSE;
   
    // update tmin and tmax of the ray 
    rRay.minmax(tTmin, tTmax);
   
    //  begin w/ the root node... 
    tCur = mRoot;
    sTraversStack.push(BSP_StkNode(0, 0, 0));

    // Iterate, util the stack not empty...
    while ( tCur )  {
      
	// Is the current node a leaf ? 
	while ( tCur->mChild[0] )  {
	  
	    // calculate the distance from the ray-origin to the subdiv.
	    // plane. This rountine also determines the 2 FAR and NEAR
	    // children. 
	    tCur->classify(rRay, tDist, tNear, tFar);
	 
	    // Only intersection w/ NEAR child possible! 
	    if ((tDist > tTmax) || (tDist <= 0) ) {
		tCur = tNear;
	    } 
	    // the same w/ FAR! 
	    else if (tDist < tTmin)   {
		tCur = tFar;
	    } 
	    // the ray can intersect w/ both children 
	    else  {
	    
		// PUSH the FAR child into the stack 
		sTraversStack.push(BSP_StkNode(tFar, tDist, tTmax));
	    
		// The next node to work with is the NEAR child ! 
		tCur   = tNear;
		tTmax  = tDist;
	    }
	}
      
	// update the tmin & tmax parameters of the ray 
	rRay.setTmin(tTmin);
	rRay.setTmax(tTmax);
      
	// intersect w/ all children of the current node. 
	if( tCur->intersectNearest(rRay, rInter)) {
	//if( tCur->intersect(rRay, rInter)) {
	    tHIT = TRUE;
	    break; 
	}
      
      
	// POP the last FAR child 
	BSP_StkNode   stkNode = sTraversStack.pop();
	tCur   = stkNode.mNode;
	tTmin  = stkNode.mMin;
	tTmax  = stkNode.mMax;
    }
    
    sTraversStack.reset();
    return tHIT;
}
 
// 
// Add the new primitives to the children list and
// expand the root-node to bound them all !
//
GT_BSP_Tree::GT_BSP_Tree(GTScene *pScene) {
    mRoot = new GT_BSP_Node;
    
    class GTConvertIter: public GIterator<GTObject>  {
     public:
	GTConvertIter(GTObjectList *pList, GT_BSP_Node* pRoot) {
	    mObjectList = pList;
	    mRoot = pRoot;
	}
	bool exec(GTObject& pObj)  {
	    mObjectList->append(&pObj);
	    mRoot->add(&pObj);
	    return (FALSE);
	}
     private:
	GTObjectList* mObjectList;
	GT_BSP_Node*  mRoot;
    } tIter(&mListObjects, mRoot);

    pScene->doWithElements( tIter);
}
 


void  GT_BSP_Tree::buildHierarchy(void) {
    mMaxListLength = 10;

    mMaxDepth = (int) ceil(log((GFloat) (mListObjects.count())/
			       (GFloat)  mMaxListLength)
			   / log(2.0));
    
    if (mMaxDepth < 0) 
      mMaxDepth = 0;
   
    cerr << "Building BSP-Tree w/ max. depth= " << mMaxDepth
      << ", max. prims in a node= " << mMaxListLength << "\n";
    
    mRoot->mAxis = X_AXIS;
    __subdivide(mRoot, 0, mMaxDepth, mMaxListLength);
   
    cerr << "Building BVH-Hierarchies for each BSP-Node...\n"; 
    __buildBVH(mRoot);
}

// Dtor
GT_BSP_Tree::~GT_BSP_Tree(void)   {
    
    // Destroy our BSP-Tree, incl. BVHs in each BSP-node!
    KILL_MEM(mRoot);

    // What ? The real primObjects still alive ? 
    // KILL them in the primary list to complete this DooMed-task!
    //for(ListItr(GTObjectList) _Itr(mListPrimObjects);
    //	_Itr.more(); _Itr.next())
    // delete _Itr.cur();
   
    // Maximmum Overkill :)
}

#ifdef GTD_DEBUG
void 
GT_BSP_Node::print(FILE *f){
   space(f, m_iLevel*2);
   s << "BSP{\n"; 
   
   if(mtChild[0]==NULL) {
      space(f, (m_iLevel+1)*2);
      s << "$" << ((BVH_Node *)&A)->Cost() << "\n";
      s << *((RTE_BBox *)&A);
      
   }else {
      A.mtChild[0]->SetLevel(A.m_iLevel+1); 
      A.mtChild[1]->SetLevel(A.m_iLevel+1); 
      s << *((BSP_Node*)(A.mtChild[0])); 
      space(s, (A.m_iLevel+1)*2); s << "/* NEAR */\n";
      
      s << *((BSP_Node *)(A.mtChild[1])); 
      space(s, (A.m_iLevel+1)*2); s << "/* FAR */\n";
   }
   space(s, A.m_iLevel*2);
   s << "}\n"; 
   
   return s;
};

#endif
