

#include <stdlib.h>
#include <time.h>

#include "gt_defs.h"
#include "gt_cam.h"

#include "gt_light.h"

GTRayTracer::~GTRayTracer(void) {
    KILL_MEM(mBSP_Tree);
}


/*
 * %FUNCT: GTRayTracer::trace_ShadowRay
 * 
 * %DESCR: trace a shadow ray and return the light attenuation.
 *         If an opaque object was found, save it in pLastOpaque
 *         for the shadow cache.
 * 
 * %NOTES: Many optimizations (and simplifications, too:) are impl.
 *         here: We save all transparent objects in a list for later
 *         testing. For the worse case, if no opaque object was found, 
 *         check this list and multiple all attenuation factors of 
 *         transparent-objects along the ray. This simplification is 
 *         neccessary for shadow-calculation, because refraction 
 *         calculation is expensive for this task. 
 */
GColorRGB GTRayTracer::traceShadowRay(GRay3&         rRay,
				      GRayTraceable* rLastOpaque)
{
    GInterSectionList tInterList;
    
    if (mBSP_Tree->intersect(rRay, tInterList)) {
	rLastOpaque = &(tInterList.getFirst().getPrimitive());
	return (GColorRGB(0, 0, 0));
    } else { 
	rLastOpaque = 0;
	return (GColorRGB(1.0, 1.0, 1.0));
    }
}


GColorRGB GTRayTracer::traceShadingRay(int       pLevel, 
				       GFloat    pWeight, 
				       GRay3&    rRay)
{
    GInterSectionList tInterList;
    GColorRGB         tColor;
    
    // find the nearest intersection, if anyone... 
    if (mBSP_Tree->intersect(rRay, tInterList))  {
      
	// shade the nearest object 
	GInterSection tInter = tInterList.getFirst();
	tColor = shade(pLevel,  pWeight, rRay, tInter);
    }
    else 
      tColor = getBackground();
   
    // clip color to [0,1]:
    tColor.correct();
   
    return tColor;
}

GColorRGB GTRayTracer::trace(int    pLEVEL, 
			     GFloat pWEIGHT, 
			     GRay3& rRay)
{
    // if(!(rRay.isShadowRay()))
    return(traceShadingRay(pLEVEL, pWEIGHT, rRay)); 
    // else
    // return(traceShadowRay());
}
 

GColorRGB GTRayTracer::shade(int    pLevel, 
			     GFloat pWeight, 
			     GRay3& rRay, 
			     const GInterSection& pInter)
{
    GVector3       tPoint(rRay.extrap(pInter.getT()));
    GRayTraceable& tPrim = pInter.getPrimitive();
    GSurface       tSurf(tPrim.getSurface(tPoint));
    GVector3       tNormal(tPrim.getNormal(tPoint));
      
   
    GVector3 tView(rRay.getVector());
    GVector3 tReflect(tView + tNormal * -2 * tNormal.dot(tView));
   
    GTShadingParams tShadeP(tPoint, tNormal, GVector3(0, 0, 0), tReflect, tView,
			    pInter.enter(), &tSurf, this);
   
    GColorRGB tColor(0, 0, 0);
   

    // ### DIFFUSE && SPECULAR 
    // compute all contributions for direct illumination 
    for (GTLightItr _Itr(mListLights); 
	 _Itr.more(); 
	 _Itr.next())  {
	
	tColor += _Itr.cur()->shade(tShadeP);
    }

   
    if(pLevel + 1 >= gMaxRecursion)
      goto L_EXIT;

    // ### REFLECTION
    // recurse on reflection ray if significant 
    if (tSurf.getSpecular().abs() * pWeight > gMinWeight)  {
	// contruct a refl. ray...
	GRay3 tReflRay(tPoint, tReflect, GRay3::R_REFLECT);
      
	// and let our camera do the hard job!
	tColor += tSurf.getSpecular() * 
	  traceShadingRay(pLevel + 1, 
			  tSurf.getSpecular().abs() * pWeight, 
			  tReflRay);
    }
   
    // ### TRANSMISSION
    if (tSurf.getTransparency() * pWeight > gMinWeight)  {
	GFloat tGamma, tDot, tSqrt;
      
	// comming from air or going into air ? 
	tDot = -tNormal.dot(tView);
	if (tDot > 0)
	  tGamma = 1 / tSurf.getRefraction();
	else  {
	    tGamma  = tSurf.getRefraction();
	    tDot    = - tDot;
	    tNormal = tNormal.negate();
	}
	tSqrt = 1.0 - tGamma * tGamma * (1.0 - tDot * tDot);
      
	// check for total internal refraction 
	if (tSqrt > 0)  {
	    tSqrt = tGamma * tDot - sqrt(tSqrt);
	    GRay3 tTransRay(tPoint, 
			    tNormal * tSqrt + tView * tGamma,
			    GRay3::R_REFRACT);
	 
	    tColor += 
	      traceShadingRay(pLevel + 1, tSurf.getTransparency() * pWeight, 
			      tTransRay) * tSurf.getTransparency();
	}
    }
   
    L_EXIT:
    return tColor;
}

void GTRayTracer::rendering(GTScene *pScene)  {
    time_t beg, end;
    
    ///// Preparing...
    time(&beg);

    // build the list of lights 
    mListLights.reset();
    class GTConvertLightsIter: public GIterator<GTLight>  {
	GTLightList *mLights;
     public:
	GTConvertLightsIter( GTLightList *pLights)  {
	    mLights = pLights; 
	}
	bool exec(GTLight& pL)  {
	    mLights->append(&pL);
	    return (false);
	}
    } tIter( &mListLights );
    gGTLights->doWithElements(tIter );
   
   
   
    // build the BSP-Tree
    KILL_MEM(mBSP_Tree);
    mBSP_Tree = new GT_BSP_Tree(pScene);
    mBSP_Tree->buildHierarchy();

    
    time(&end);
    cerr << "Preparing time:\t" <<  end -beg << "\n";
	
    ///// rendering...
    int pw = mRealization->getWidth();
    int ph = mRealization->getHeight();
    GFloat wc = 1.0/(pw - 1);
    GFloat hc = 1.0/(ph - 1);
    
    GFloat    _xx, _yy;
	
    GFloat tAspect = (GFloat)pw / (GFloat)ph;
    GFloat tViewDistance  = 1/tan(getZoom()/2.0 * M_PI/180.0);
    
    time(&beg);
    
    register int i;
    register int j;
    for (j = ph-1, _yy = 1.0; j >= 0; j--, _yy = _yy - hc) {
	for (i = 0, _xx = 0.0; i < pw; i++, _xx = _xx + wc) {
	    // contruct an eye ray, caching is possible! 
	    GRay3  tRay(getViewpoint(), 
			(getMatrix() * GVector3((1 - _xx - _xx) * tAspect, 
						-1 + _yy + _yy, 
						tViewDistance) 
			 - getViewpoint()).unitize(), 
			GRay3::R_EYE);
	    mRealization->putPixel( i, j, traceShadingRay( 0, 1, tRay));
	}
    }
    
    // GLookatRayTracer<GTObject, GPixmapDisplay>::rendering(*pScene);
    
    
    
    time(&end);
    cerr << "Raytracing time:\t" << end -beg << "\n";
}
 


