#include "render/graytrac.h"
#include "gt_obj.h"

enum { X_AXIS=0, Y_AXIS=1, Z_AXIS=2 };


GTPolygon::GTPolygon(int pNumVerts,	
		     GPoint3 *pVertices)
: GTPrimitive()
{
    mNumVerts = pNumVerts;		  
    mVertices = new GPoint3[mNumVerts];
    for (register int i = 0; i < mNumVerts; i++)
      mVertices[i] = pVertices[i];
    
    // compute a bounding box for BVH  
    computeABox();
    // compute the plane equation, the dominant axis and the projection 
    // rectangle along this axis.
    precompute();
}

GTPolygon::GTPolygon(const GVertexAggregate& pVertices) 
: GTPrimitive() 
{
    mNumVerts = pVertices.getNumber();		  
    mVertices = new GPoint3[mNumVerts];

    for(register int i = 0; i< mNumVerts; i++) 
      mVertices[i] = (pVertices[i]).getPoint();
    
    // compute a bounding box for BVH  
    computeABox();
    // compute the plane equation, the dominant axis and the projection 
    // rectangle along this axis.
    precompute();

}

GTPolygon::~GTPolygon(void)
{
    KILL_ARR(mVertices);
}

void GTPolygon::precompute(void)
{
    //// Compute the plane equation of the polygon.
    GVector3 tX((mVertices[2] - mVertices[1]).cross(mVertices[0] - mVertices[1]));
    //GVector3 tX((mVertices[0] - mVertices[1]).cross(mVertices[2] - mVertices[1]));
    if ( tX.abs() > gEpsilon) {
	mValid = TRUE;
	mNormal = tX.unitize();
    }
    else {
	mValid = FALSE;
	return;
    }
    mDist = mNormal.dot(mVertices[0]);

    //// Determine the dominant axis of the normal
    mDominantAxis = X_AXIS;		
    GFloat tMax = fabs(mNormal.getX());
    if (fabs(mNormal.getY()) > tMax) { 
	tMax = fabs(mNormal.getY());	  
	mDominantAxis = Y_AXIS;
    }
    if (fabs(mNormal.getZ()) > tMax)	
      mDominantAxis = Z_AXIS;		

    mVertices[0].project(mUmin, mVmin, mDominantAxis);
    mUmax = mUmin;
    mVmax = mVmin;
    for (register int i = 1; i < mNumVerts; i++) {
	GFloat tX, tY;
	mVertices[i].project(tX, tY, mDominantAxis);
	if (tX < mUmin)
	  mUmin = tX;
	else if (tX > mUmax)
	  mUmax = tX;
	if (tY < mVmin)
	  mVmin = tY;
	else if (tY > mVmax)
	  mVmax = tY;
    }
}


static bool __intersectSegment(GFloat pX,  GFloat pY, 
			       GFloat pU1, GFloat pV1, 
			       GFloat pU2, GFloat pV2)
{

    pU1 -= pX; pU2 -= pX;
    pV1 -= pY; pV2 -= pY;

    //// check the intervall
    if ((pV1 > 0 && pV2 > 0) ||
	(pV1 < 0 && pV2 < 0) ||
	(pU1 < 0 && pU2 < 0))
      return FALSE;

    if (pU1 > 0 && pU2 > 0)
	return 1;

    GFloat tYdiff = pV2 - pV1;
    if (fabs(tYdiff) < gEpsilon ) {	
	if (((fabs(pV1) > gEpsilon) ||
	     (pU1 > 0) || (pU2 > 0)))
	  return FALSE;
	return TRUE;
    }

    return ((pU1 + pV1  * (pU1 - pU2) / tYdiff) > 0);
}





// Polygon intersection routine.
bool GTPolygon::intersect(const GRay3& pRay, GInterSectionList& rInter) 
//int Polygon::NearestInt(Ray& ray, float& t, float maxt)
{

    //// intersect w/ the plane
    if (!mValid)
      return FALSE;

    GFloat tDot = mNormal.dot(pRay.getVector());
    if (fabs(tDot) < gEpsilon) 
      return FALSE;

    //// compute the t-parameter
    GFloat tT = (mDist - mNormal.dot(pRay.getPoint())) / tDot;
    if(!pRay.isValid(tT)) 
      return FALSE;
    
    GFloat tU, tV;
    pRay.extrap(tT).project(tU, tV, mDominantAxis);

    if (tU < mUmin || tU > mUmax || tV < mVmin || tV > mVmax)
      return FALSE;

    int tCount = 0;
    GFloat tU1, tV1, tU2, tV2;

    mVertices[mNumVerts - 1].project(tU1, tV1, mDominantAxis);
    
    register GPoint3 *tVertexPtr = mVertices;
    for (register i = 0; i < mNumVerts; i++, tVertexPtr++) {
	tVertexPtr->project(tU2, tV2, mDominantAxis);
	tCount += __intersectSegment(tU, tV, tU1, tV1, tU2, tV2);
	
	tU1 = tU2;
	tV1 = tV2;
    }
    if(tCount & 1) {
	rInter.add( GInterSection( self, tT, 1));
	return TRUE;
    }
    else
      return FALSE;
}



bool GTSphere::intersect(const GRay3& pRay, GInterSectionList& rInter) {
    // origin to center 
    GVector3 _o2c(mCenter - pRay.getPoint());      
    // square of tangential distance
    GFloat _tan_sq = _o2c.dot(_o2c) - mRad * mRad;
   
    // projected distance of _o2c to rays direction  
    GFloat _project = _o2c.dot(pRay.getVector());
   
    // if the rays origin is outside and the rays direction
    // points away to the sphere, no intersection occurs.
    if( _tan_sq >= 0 && _project < gEpsilon )  
	return false;
   
    GFloat _d = _project * _project - _tan_sq;
    
    if (_d < 0) return false;
   
    _d = sqrt(_d);
    GFloat _tmax = _project + _d;
    GFloat _tmin = _project - _d;
    register int _HIT = false;
    if (pRay.isValid(_tmax)) {
	rInter.add( GInterSection( self, _tmax, 0));
	_HIT = true;
    }
    if(pRay.isValid(_tmin)) {
	rInter.add( GInterSection( self, _tmin, 1));
	_HIT = true;
    }
    return (_HIT);
}
    

