#ifndef __GT_OBJ_H__
#define __GT_OBJ_H__

#include "gt_defs.h"
#include "gt_abox.h"

class GTObject: public GRayTraceable {
    // CONCEPT: Derived test graphics kernel
    //
    // ABSTRACT PURPOSE: Base class for GT objects  
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
 public:
    enum {
	O_PRIMITIVE    = 0x0,
	O_BOUND        = 0x1,
	O_CSG          = 0x2,
	O_GROUP        = 0x4
    };
    GTObject(int F = O_PRIMITIVE): mFlags(F) {}
    virtual ~GTObject(void) {}
    int flags(void) const { return(mFlags); }
    
    bool isBound(void) {
	return (mFlags & O_BOUND);
    }
    bool isCSG(void) {
	return(mFlags & O_CSG);
    }
    bool isGroup(void) {
	return(mFlags & O_GROUP);
    }
   
    virtual int getNumOfChildren(void) const { return(0);}
    
    virtual GFloat getCost(void) const = 0;
    virtual bool isOverlap(const GPoint3& point) const = 0;
    virtual bool isOverlap(const GTABox& box) const = 0;
    virtual GTABox getABox(void) const = 0;  
 
 protected:
    int mFlags;	 
}; 



class GTPrimitive: public GTObject {
    // CONCEPT: Derived test graphics kernel
    //
    // ABSTRACT PURPOSE: Abstract class for GT primitives
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
 public:
    GTPrimitive(GSurface *pSurf = 0)
      : GTObject(GTObject::O_PRIMITIVE),
      mSurface(pSurf) {
	  // computeABox();
      }
   
    virtual ~GTPrimitive(void) {}
    virtual GTABox  getABox(void) const { 
	return(mBox); 
    }  
    
    virtual GFloat getCost(void) const{ return(0); }
    virtual bool isOverlap(const GPoint3& point) const {
	return (mBox.isOverlap(point));
    };
    virtual bool isOverlap(const GTABox&    box) const {
	return (mBox.isOverlap(box));
    };

    // get surface at specified WC point:
    virtual const GSurface& getSurface(const GPoint3&) {
	return((mSurface) ? *mSurface : gDummySurface);
    }
    virtual void setSurface(const GSurface *pSurface) {
	mSurface = pSurface;
    }
  
 protected:
    GTABox     mBox;
    GSurface  *mSurface;

    virtual void computeABox(void) {
	mBox = GTABox(GPoint3(-MAXFLT, -MAXFLT, -MAXFLT), 
		      GPoint3(MAXFLT, MAXFLT, MAXFLT));
    }

};

typedef GDLList<GTObject> GTScene;

class GTSphere: public GTPrimitive {
    // CONCEPT: Derived test graphics kernel
    //
    // ABSTRACT PURPOSE:   
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 

 public:
    GTSphere(const GPoint3& pCenter = GPoint3(0,0,0), GFloat pRad = 1.0)
      : mCenter(pCenter), 
      mRad(pRad), 
      GTPrimitive() { computeABox(); }
    
    bool intersect(const GRay3&, GInterSectionList&);
    GVector3 getNormal(const GPoint3& rPoint) {
	return((rPoint - mCenter).unitize());
    }
    void setPos(const GPoint3& pPoint) { 
	mCenter = pPoint;
	computeABox();
    }
    void setRad(GFloat pRad) { 
	mRad = pRad;
	computeABox();
    }
    
 private:
    GPoint3 mCenter;
    GFloat mRad;
    virtual void computeABox(void) {
	mBox = GTABox(mCenter - GVector3(mRad, mRad, mRad),
		      mCenter + GVector3(mRad, mRad, mRad));
    }
};



class GTCompoPolygon: public GTPrimitive {
    // CONCEPT: Derived test graphics kernel
    //
    // ABSTRACT PURPOSE: A class for planar polygons    
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: 
    //
    // CONSTANTS: 
 public:
    // Create a polygon from a const list of vertices:
    GTCompoPolygon(const GVertexAggregate& pVertices) 
      : GTPrimitive() {
	  mPoly = new GRayTraceablePolygon(pVertices);
	  mPoly->geometryChanged();
	  computeABox();
      }
    
    virtual ~GTCompoPolygon(void) { 
	KILL_MEM(mPoly);
    }
    
    bool intersect(const GRay3& pRay, GInterSectionList& rInter) {
	GInterSectionList tInter;
	if(mPoly->intersect(pRay, tInter)) {
	    rInter.add( GInterSection(self, 
				      tInter.getFirst().getT(), 
				      tInter.getFirst().enter()));
	}
    }
    GVector3 getNormal(const GPoint3& rPoint) {
	return (mPoly->getNormal(rPoint));
    }

    
    
 private:
    GRayTraceablePolygon *mPoly;
    
    virtual void computeABox(void) {
	const GVertexList& tVerts = mPoly->getVertexList();
	mBox = GTABox(GPoint3(MAXFLT, MAXFLT, MAXFLT), 
		      GPoint3(-MAXFLT, -MAXFLT, -MAXFLT));
	for(register int i = 0; i< tVerts.getNumber(); i++) 
	  mBox += (tVerts[i]).getPoint();
		
	GVector3 tDiff(mBox.min() - mBox.max());
	for (i = 0; i < 3; i++)
	  if (fabs(tDiff[i]) < gEpsilon) {
	      mBox.min()[i] -= gEpsilon;
	      mBox.max()[i] += gEpsilon;
	  }
    }
};



class GTPolygon : public GTPrimitive {
    // CONCEPT: Derived test graphics kernel
    //
    // ABSTRACT PURPOSE: A class for planar polygons (direct)   
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: 
    //
    // CONSTANTS: 
 public:
    GTPolygon(void)
      : GTPrimitive(),
      mNumVerts(0), mVertices(0) { }
    
    GTPolygon(int pNumVerts, GPoint3 *pVertices);
    GTPolygon(const GVertexAggregate& pVertices) ;
    virtual ~GTPolygon();

    bool intersect(const GRay3&, GInterSectionList&);
    GVector3 getNormal(const GPoint3&) {
	return (mNormal);
    }
    
 protected:
    int mNumVerts;		
    GPoint3 *mVertices;	

    GVector3 mNormal;		
    GFloat mDist;			
    bool mValid;			
				
    int mDominantAxis;			
    GFloat mUmin, mUmax;		
    GFloat mVmin, mVmax;		

    void precompute(void);		

        
    virtual void computeABox(void) {
	mBox = GTABox(GPoint3(MAXFLT, MAXFLT, MAXFLT), 
		      GPoint3(-MAXFLT, -MAXFLT, -MAXFLT));
	for (register int i = 0; i < mNumVerts; i++)
	  mBox += mVertices[i];
	
	GVector3 tDiff(mBox.min() - mBox.max());
	for (i = 0; i < 3; i++)
	  if (fabs(tDiff[i]) < gEpsilon) {
	      mBox.min()[i] -= gEpsilon;
	      mBox.max()[i] += gEpsilon;
	}
    }
};







#endif
