#include <fstream.h>
#include <time.h>
#include "render/graycam.h"
#include "render/gpixmap.h"

#include "gt_obj.h"
#include "gt_light.h"
#include "gt_cam.h"

void Read_NFF(const char *file_, GTRayTracer* rRT, GTScene* rSc);

static void prerror(const char *err) {
   cerr << err << '\n';
   exit(3);
}

int main(int argc, char *argv[]) {
    if(argc<2)
      prerror("Usage: nff NFF-file");

    gInitialize();
   
    GTRayTracer  aRT;
    GTScene aSc(true);
    gGTLights  = new GDLList<GTLight>(true);
    Read_NFF(argv[1], &aRT, &aSc);
    aRT.setSamples(1); 
    
    aRT.rendering(&aSc);
    
    return (0);
}


void Read_NFF(const char *file_, GTRayTracer* rRT, GTScene* rSc){
    static char nexttok[256];
    char cmd;
    ifstream inpf(file_);
   
    if (! inpf) {
	sprintf(nexttok, "Could not open %s", file_);
	prerror(nexttok);
    }
    nexttok[0] = '\0';

    GSurface      *tSurf = 0;

    while (! inpf.eof()) {
	if (nexttok[0]) {
	    cmd = nexttok[0];
	    nexttok[0] = '\0';
	}
	else {
	    inpf >> cmd;
	    if (! inpf)
	      break;
	}
	
	switch (cmd) {

	 case 'v':  
	    //// Camera
	      
	    GFloat fromx, fromy, fromz;
	    GFloat atx, aty, atz;
	    GFloat upx, upy, upz;
	    GFloat ang, hither;
	    int xres, yres;
	      
	    inpf >> nexttok;
	    if (strcmp(nexttok, "from"))
	      prerror("\"from\" expected");
	    inpf >> fromx >> fromy >> fromz >> nexttok;
	 
	    if (strcmp(nexttok, "at"))
	      prerror("\"at\" expected");
	    inpf >> atx >> aty >> atz >> nexttok;

	    if (strcmp(nexttok, "up"))
	      prerror("\"up\" expected");
	    inpf >> upx >> upy >> upz >> nexttok;

	    if (strcmp(nexttok, "angle"))
	      prerror("\"angle\" expected");
	    inpf >> ang >> nexttok;
	 
	    if (strcmp(nexttok, "hither"))
	      prerror("\"hither\" expected");
	    inpf >> hither >> nexttok;
	    if (strcmp(nexttok, "resolution"))
	      prerror("\"resolution\" expected");
	    inpf >> xres >> yres;
	    nexttok[0] = '\0';
	    
	    GPoint3 tViewer(fromx, fromy, fromz); 
	    GPoint3 tLookAt(atx, aty, atz);
	    
	    rRT->setViewpoint(tViewer);
	    rRT->setRefpoint(tLookAt);
	    rRT->setZoom(ang);
	    
	    GVector3 tUp(upx, upy, upz);
	    tUp = tUp.unitize();
	    
	    GVector3 tN(tViewer - tLookAt);
	    tN = tN.unitize();

	    GVector3 ye(0, 1, 0);
	    // defaults
    
	    if(tN == GVector3(0, 1, 0))
	      ye = GVector3(0, 0, 1);
	    if(tN == GVector3(0, -1, 0))
	      ye = GVector3(0, 0, -1);
	    
	    GVector3 xe(ye.cross(tN)); 
	    xe = xe.unitize();
	    ye = tN.cross(xe); 
 
	    GVector3 tV(tUp - tLookAt);
	    tV -= tN * (tV.dot(tN));
	    
	    if(tV == GVector3(0, 0, 0))
	      tV = ye;
	    else
	      tV = tV.unitize();
	    
	    rRT->setTwist(M_R2D * acos(tV.dot(ye)));
	      
	    GPixmapDisplay  *aPix = new GPixmapDisplay(file_, xres, yres);
	    rRT->initRealization(*aPix);
	    
	    break;
	 
	 case 'b':
	    //// Background
	    
	    GFloat x, y, z;
	    inpf >> x >> y >> z;
	    rRT->setBackground(GColorRGB(x, y, z));
	    break;

	 case 'l':
	    //// Light
	    
	    GFloat lx, ly, lz;
	    GFloat cx, cy, cz;
	    char *sc;
	    
	    inpf >> lx >> ly >> lz;
	    inpf >> nexttok;
	    cx = strtod(nexttok, &sc);
	    
	    GTPointLight *_pL = new GTPointLight();
	    _pL->setPosition(GPoint3(lx, ly, lz));
	    if (sc) {
		_pL->setColor(GColorRGB(1,1,1));
	    }else {
		inpf >> cy >> cz;
		_pL->setColor(GColorRGB(cx, cy, cz));
		nexttok[0] = '\0';
	    }
	    
	    gGTLights->append(_pL);
	    break;

       case 'f':
	    //// Surface
	    
	      {  GFloat r, g, b, kd, ks, shine, t, ior;
		  inpf >> r >> g >> b >> kd >> ks >> shine >> t >> ior;
		  tSurf = new GSurface();
		  tSurf->setAmbient(GColorRGB(r, g, b));
		  tSurf->setDiffuse(GColorRGB(kd*r, kd*g, kd*b));
		  tSurf->setSpecular(GColorRGB(ks*r, ks*g, ks*b));
		  tSurf->setShininess(shine);
		  tSurf->setTransparency(t);
		  tSurf->setRefraction(ior);
	      }
	    break;
	    
	 
       case 's':
	    //// sphere
	    
	      {  GFloat x, y, z, r;
		  inpf >> x >> y >> z >> r;
		  GTSphere *_pSph = new GTSphere(GPoint3(x, y, z), r);
		  _pSph->setSurface((tSurf) ? tSurf : &gDummySurface);
		  rSc->append(_pSph);
	      }
	    break;
	 
	 case 'p':
	    if (inpf.peek() == 'p') {   
		
		// polygon patch( $Phong!)
		inpf >> cmd;

#ifdef GTD_COMPOPOLY
		int n;
		inpf >> n;
		GVertexList *tVerts = new GVertexList(); 
		// TRUE or FALSE !?
		
		for (int i = 0; i < n; i++) {
		    GFloat x, y, z, x1, y1, z1;
		    inpf >> x >> y >> z >> x1 >> y1 >> z1;
		    tVerts->append(new GVertex(GVector3(x, y, z), 
					       GVector3(x1, y1, z1)));
		}
		GTCompoPolygon *tPoly = new GTCompoPolygon(*tVerts);
		tPoly->setSurface((tSurf) ? tSurf : &gDummySurface);
		rSc->append(tPoly);
		delete tVerts;
#else
		int tNum;
		inpf >> tNum;
		GPoint3 *tVerts = new GPoint3[tNum];
		for (register int i = 0; i < tNum ; i++) {
		    GFloat x, y, z, nx, ny, nz;
		    inpf >> x >> y >> z >> nx >> ny >> nz;
		    tVerts[i] = GVector3(x, y, z);
		}
		GTPolygon *tPoly = new GTPolygon(tNum, tVerts);
		tPoly->setSurface((tSurf) ? tSurf : &gDummySurface);
		rSc->append(tPoly);
		delete [] tVerts;
#endif
	    }
	    else {		
#ifdef GTD_COMPOPOLY
		// planare GTCompoPolygons
		int n;
		inpf >> n;

		GVertexList tVerts(false);
		for (register int i = 0; i < n ; i++) {
		    GFloat x, y, z;
		    inpf >> x >> y >> z;
		    tVerts.append(new GVertex(GVector3(x, y, z)));
		}
		GTCompoPolygon *tPoly = new GTCompoPolygon(tVerts);
		tPoly->setSurface((tSurf) ? tSurf : &gDummySurface);
		rSc->append(tPoly);
#else
		int tNum;
		inpf >> tNum;
		GPoint3 *tVerts = new GPoint3[tNum];
		for (register int i = 0; i < tNum ; i++) {
		    GFloat x, y, z;
		    inpf >> x >> y >> z;
		    tVerts[i] = GVector3(x, y, z);
		}
		GTPolygon *tPoly = new GTPolygon(tNum, tVerts);
		tPoly->setSurface((tSurf) ? tSurf : &gDummySurface);
		rSc->append(tPoly);
		delete [] tVerts;
#endif
	    }
	    break;
	 
	 case '#':
	    inpf.getline(nexttok, 256);
	    nexttok[0] = '\0';
	    break;
	}
    }

}
