#ifndef G_DEVICE_H
#define G_DEVICE_H

#include "input/gevent.h"
#include "input/gcallbck.h"
#include "render/grealiz.h"
#include "util/gstring.h"
#include "util/gtempl.h"

class GInputDevice {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Base class of input devices.
    // An input device may be used as child of another 
    // input device OR may be used stand-alone.
    // It may be bound to a realization, too.
    // INVARIANT: 
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: gNoInputDevice
  public:
    GInputDevice(GInputDevice*);
    virtual ~GInputDevice(void);

    virtual void on(void) { mState = true; }
    // turn on the device
    virtual void off(void) { mState = false; }
    // turn off the device
    bool isOn(void) const { return(mState); }

    // get the children of the device:
    const GDLList<GInputDevice> &getChildren(void) const { 
	return(cldList); 
    }

    const GTime& getEventTime(void) const {
	return(mTime); 
    }

    // add a C++ callback to the list:
    void addCB(GCallback* fun) { 
	cbList.append(fun);
    }

    // remove a C++ callback from list:
    void removeCB(GCallback* fun) { 
	cbList.remove(fun);
    }

    // The default event method for the input devices.
    // Send the event to the children.
    // May be used in subclasses.
    // Return true if the object was triggered.
    virtual bool event(GEvent* pEvent = 0) {
	return(callChildren(pEvent));
    }

    GInputDevice* getFather(void) {
	return(mFather); 
    }

    void setRealization(GRealization* pReal) { mReal = pReal; }

    GRealization* getRealization(void) {
	if (mReal) return(mReal);
	else {
	    if (getFather()) return(getFather()->getRealization());
	    else return(0);
	}
    }
    
    // Return true if the object is DIRECTLY assigned
    // to a realization. In that case it may ask the 
    // realization for events.
    bool isRoot(void) {
	return(mReal ? true: false);
    }
  protected:
    void addChild(GInputDevice*);
    void removeChild(GInputDevice*);
    
    void callCBs(void);
    // call all callbacks in the callback list
    bool callChildren(GEvent*);
    // call all child objects 
    void setTime(const GTime& pTime) {
	mTime = pTime;
    }
  protected:
    // time of last triggering:
    GTime mTime;
  private:
    // the related father object:
    GInputDevice* mFather;

    bool mState;
    // The state of the input device. 
    // True means that the device is on.
    
    GDLList<GInputDevice> cldList;
    // the child list 

    GDLList<GCallback> cbList;
    // the callback list

    GRealization* mReal;
};

inline void GInputDevice::addChild(GInputDevice* d) { cldList.append(d); }
inline void GInputDevice::removeChild(GInputDevice* d) { cldList.remove(d);  }

extern const GInputDevice* gNoInputDevice;

template<class I>
class GPollInputDeviceIterator: public GIterator<I> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An iterator that calls the
    // event method.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GPollInputDeviceIterator(GEvent* pEvent): mEvent(pEvent), mWasEvent(false) {}
    virtual bool exec(I& i) {
	if (i.event(mEvent)) mWasEvent = true;
	return(false);
    }
    bool sleep(void) {
	bool ret = !mWasEvent;
	mWasEvent = false;
	return(ret);
    }
  private:
    GEvent* mEvent;
    bool mWasEvent;
}; 

#endif

