#ifndef G_EVENT_H
#define G_EVENT_H

#include <string.h>
#include "input/gtime.h"

class GEvent {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Basic class of classes
    // representing events.
    // INVARIANT: An event has a unique type and
    // a creation time.
    // IMPLEMENTATION ISSUES:  
    //
    // CONSTANTS: 
    // 
  protected:
    GEvent(void) {}
  public:
    // inquire the class:
    virtual const char *getClass(void) const = 0;

    // Check type identity. The method must be
    // overloaded only for non-directly derived
    // event types.
    virtual bool isA(const char* pClass) {
	if (pClass == getClass()) return(true);
	else return(!strcmp(pClass, getClass()));
    }
    const GTime& getTime(void) const { return(mTime); }
  private:
    GTime mTime;
};

class GMouseEvent: public GEvent {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: This event type represents
    // mouse actions and the state of some special keys.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: (0, 0) is the lower left
    // window corner.
    // CONSTANTS: 
  public:
    GMouseEvent(void): left(false), middle(false), right(false),
    shift(false), ctrl(false), alt(false),
    x(0), y(0), w(0), h(0) {}
  public:
    // state of the mouse buttons (true if pressed):
    bool left, middle, right;
    // state of the special keys (true if pressed):
    bool shift, ctrl, alt;
    // mouse pointer coordinates and pixmap dimensions:
    int x, y, w, h;
};

extern const char* gMouseButtonEventID;

class GMouseButtonEvent: public GMouseEvent {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: This event type represents
    // mouse press/release actions and the state of some
    // special keys.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: 
    // 
    // CONSTANTS: 
  public:
    virtual const char* getClass(void) const { return(gMouseButtonEventID); }
};

extern const char* gMouseMotionEventID;

class GMouseMotionEvent: public GMouseEvent {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: This event type represents
    // mouse motions and the state of some
    // special keys.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: 
    // 
    // CONSTANTS: 
  public:
    virtual const char* getClass(void) const { return(gMouseMotionEventID); }
};

extern const char* gSpaceballEventID;

class GSpaceballEvent: public GEvent {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: This event type represents
    // incremental spaceball actions.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GSpaceballEvent(void): transX(0), transY(0), transZ(0),
    twistX(0), twistY(0), twistZ(0),
	b1(0), b2(0), b3(0), b4(0), b5(0), b6(0), b7(0), b8(0), 
	pick(0) {}
    const char* getClass(void) const { return(gSpaceballEventID); }
  public:
    // the translation values of the spaceball:
    short transX, transY, transZ;
    // the rotation values of the spaceball:
    short twistX, twistY, twistZ;
    // the button states:
    bool b1, b2, b3, b4, b5, b6, b7, b8;
    // the state of the pickbutton:
    bool pick;
    // the number of periods of 0.25 ms since sending the last non-0
    // set of Spaceball data:
    short period;
};

extern const char* gPixmapUpdateEventID;

class GPixmapDisplay;

class GUpdatePixmapEvent: public GEvent {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: This event should be generated
    // by pixmaps to force an update.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GUpdatePixmapEvent(void): partial(false), pixmap(0) {}
    virtual const char* getClass(void) const { return(gPixmapUpdateEventID); }
  public:
    bool partial;
    // if true the following coordinates are valid:  
    // (x0,y0) left upper corner of the area that must be updated
    // (x1,y1) right lower corner 
    int x0, y0, x1, y1;
    GPixmapDisplay* pixmap; 
    // the pixmap to update
};

#endif

