#ifndef G_RAY_TRACING
#define G_RAY_TRACING

#include "util/gmatrix.h"
#include "util/gsurface.h"
#include "util/gtopolgy.h"
#include "render/grender.h"

class GRayTraceable;

class GInterSection {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A data type for
    // holding material properties and other
    // data of an intersection point.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GInterSection(void): mPrim(0), mT(0), mEnter(false) {}
    GInterSection(GRayTraceable& pPrim, GFloat pT, bool pEnter): mPrim(&pPrim), mT(pT), mEnter(pEnter) {}

    GRayTraceable& getPrimitive(void) const { 
	return(*mPrim); 
    }
    GFloat getT(void) const { 
	return(mT); 
    }
    bool enter(void) const {
	return(mEnter);
    }

    GInterSection& operator=(const GInterSection& pInter) {
	mPrim = pInter.mPrim;
	mT = pInter.mT;
	mEnter = pInter.mEnter;
	return(*this);
    }

    // compare two intersections:
    bool operator < (const GInterSection&) const;
  private:
    // the hitted primitive:
    GRayTraceable* mPrim;

    // the distance from ray origin to intersection point:
    GFloat mT; 
    // true if entering the primitive, false else
    bool mEnter; 
};

class GInterSectionList {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An aggregate holding
    // intersection points.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: This intersection list stores
    // just the nearest point.
    // CONSTANTS: 
  public:
    GInterSectionList(void): mHit(false), mTransmission(1.0) {}

    // return first intersection:
    const GInterSection& getFirst(void) const {
	return(mInter);
    }

    // add intersection:
    void add(const GInterSection& s);

    int getNumber(void) const {
	return(mHit ? 1 : 0);
    }
    // return one or zero

    GFloat getTransmission(void) const {
	return(mTransmission);
    }
  private:
    bool mHit;
    GInterSection mInter;

    GFloat mTransmission;
    // A scalar in the range 0 to 1. 1 means
    // total transmission, 0 stands for no transmission.
    // The variable contains the lowest transmission found.
    // NOTE: This not a correct but a somehow useful way
    // to let shine light thru unopaque primitives.
};

template <class P>
class GRayTracingIterator: public GIterator<P> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Basic iterator
    // for ray-tracing.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: 
    // 
    // CONSTANTS: 
  public:
    GRayTracingIterator(const GRay3& pRay, GInterSectionList& pList): mRay(pRay), mList(pList), mHit(false) {}
    virtual bool exec(P& p) { 
	if (p.intersect(mRay, mList)) mHit = true;
	return(false);
    }
    void reset(void) {
	mHit = false;
    }
    // METHOD: Reset hit flag.
    bool getHit(void) const { return(mHit); }
    // METHOD: Return true if there was a hit.
  private:
    const GRay3& mRay;
    GInterSectionList& mList;
    bool mHit;
};

extern const GSurface gDummySurface;

class GRayTraceable {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Basic renderable for the
    // ray-tracer.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: gDummySurface - a surface used for the
    // get surface method. This method should never called.
  public:
    // Get nearest sub renderable by tracing
    // the ray or 0.
    // Return the distance to the entry point:
    virtual GFloat getNearest(const GRay3&, GRayTraceable*& rRay) {
	rRay = 0;
	return(0.0);
    }

    virtual bool intersect(const GRay3&, GInterSectionList& list) = 0;
    // METHOD: Add intersections to $$list$$, return true if there was a hit.

    virtual bool isHit(const GRay3& pRay) {
	GInterSectionList tList;
	return(intersect(pRay, tList));
    }
    
    
    // get normal vector at specified WC point:
    virtual GVector3 getNormal(const GPoint3& POINT) = 0;

    // get surface at specified WC point:
    virtual const GSurface& getSurface(const GPoint3&) {
	return(gDummySurface);
    }

    // Return 0 for no transmission or 1 for opaque
    // objects or an intermediate value.
    virtual GFloat getTransmission(void) const {
	return(0.0);
    }

    // bounding box in the local coordinates system:
    const GBounds3& getBounds(void) const { return(mBounds); }
  protected:
    // a set method:
    GBounds3& setBounds(void) { return(mBounds); }
  private:
    GBounds3 mBounds;
};

class GRayTraceableQuadric: public GRayTraceable {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A utility primitive for
    // ray-tracing. To be parameterized in sub-classes.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  protected: 
    // initialize quadric coefficients:
    // (quad, mixed, linear, const)
    GRayTraceableQuadric(const GVector3&, const GVector3&, const GVector3&, GFloat);
  public:
    // incremental transformations on the quadric:
    void translate(const GVector3&);
    void rotate(const GVector3&);
    void scale(const GVector3&);
  public:
    virtual GFloat getNearest(const GRay3&, GRayTraceable*&);
    virtual bool intersect(const GRay3&, GInterSectionList&);
    virtual GVector3 getNormal(const GPoint3&);
  private:
    void matrix2coefficients(void);
    void coefficients2matrix(void);
  private:
    GVector3 mQuadric, mMixed, mLinear;
    GFloat mConst;
    GMatrix4 mMat;
};

class GRayTraceableSphere: public GRayTraceableQuadric {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An ideal sphere basing on the 
    // above quadric. 
    // INVARIANT: The radius is 1.
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GRayTraceableSphere(void): GRayTraceableQuadric(GVector3(1.0, 1.0, 1.0), GVector3(0.0, 0.0, 0.0), GVector3(0.0, 0.0, 0.0), -1.0) {
	setBounds().setMin(GVector3(-1.0, -1.0, -1.0));
	setBounds().setMax(GVector3(1.0, 1.0, 1.0));
    }
};

class GRayTraceableBounds: public GRayTraceable {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A rectangular bounding box.
    // 
    // INVARIANT: 
    //
    // IMPLEMENTATION ISSUES: The dimensions have to be set
    // on the fly for a created bounding object.
    // CONSTANTS: 
    friend ostream& operator<<(ostream&, const GRayTraceableBounds&);
  public:
    // setup dimensions:
    // (min, max)
    void set(GFloat, GFloat, GFloat, GFloat, GFloat, GFloat); 
    virtual GFloat getNearest(const GRay3&, GRayTraceable*&);

    void setMin(const GVector3& pMin) { setBounds().setMin(pMin); } 
    const GVector3& getMin(void) const { return(getBounds().getMin()); }

    void setMax(const GVector3& pMax) { setBounds().setMax(pMax); } 
    const GVector3& getMax(void) const { return(getBounds().getMax()); }
  public:
    virtual bool intersect(const GRay3&, GInterSectionList&);
    virtual GVector3 getNormal(const GPoint3&);
};

typedef bool (*GIFuncPtr)(const GPoint3&, const GPoint3&, const GPoint3&);

class GRayTraceablePolygon: public GRayTraceable, virtual public GPolygon {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A planar polygon without support of vertex colors. 
    // 
    // INVARIANT: The polygon must be planar.
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    // Create an empty polygon:
    GRayTraceablePolygon(void);
    // Create a polygon from a const list of vertices:
    GRayTraceablePolygon(const GVertexAggregate&);
    virtual ~GRayTraceablePolygon(void);
    
    virtual void geometryChanged(void);

    virtual bool intersect(const GRay3&, GInterSectionList&);
    virtual GVector3 getNormal(const GPoint3&) {
	GVector3 tNormal(GPolygon::getNormal());
	return(tNormal);
    }  
  private:
    GIFuncPtr ifunc;

    GPoint3* mPolygon;
    int mPolySegs;
    bool mPolyValid;
    GFloat mDist;
    // distance of plane
};


#endif

