#ifndef G_SHADING_H
#define G_SHADING_H

#include "util/gmatrix.h"
#include "util/gsurface.h"
#include "util/gtopolgy.h"
#include "render/grender.h"
#include "render/glight.h"

// request unique number:
int gGetID(void);
// IMPLEMENTED PLATFORM DEPENDENT

// free number:
void gFreeID(int);
// IMPLEMENTED PLATFORM DEPENDENT

class GShadable {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: The base class of all primitives that
    // may be displayed via classic shading method.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: Most of methods have to be
    // implemented platform dependent. 
    // CONSTANTS: 
  protected:
    GShadable(void): mID(0) {}
  public:
    virtual ~GShadable(void);
    // IMPLEMENTED PLATFORM DEPENDENT
    virtual void shade(void) = 0;

    void callDisplayList(void);
    // IMPLEMENTED PLATFORM DEPENDENT

    // Get nearest sub renderable by tracing
    // the ray or 0.
    // Return the distance to the entry point:
    virtual GFloat getNearest(const GRay3&, GShadable*& rShad) {
	rShad = 0;
	return(0.0);
    }
  public:
    static bool fillstyle;
    // Instead of using a fillstyle attribute, that fillstyle is
    // specified via this global boolean. Primitives should
    // support both styles.
  protected:
    void setMatrix(const GMatrix4&);
    // IMPLEMENTED PLATFORM DEPENDENT
    void unsetMatrix(void);
    // IMPLEMENTED PLATFORM DEPENDENT
    void setSurface(const GSurface&);
    // IMPLEMENTED PLATFORM DEPENDENT
    void openDisplayList(void);
    // IMPLEMENTED PLATFORM DEPENDENT
    void closeDisplayList(void);
    // IMPLEMENTED PLATFORM DEPENDENT

    int getID(void) const { return(mID); }
  private:
    // the display list number:
    int mID;
};

class GShadableQuadmesh: public GShadable, virtual public GTopology2D {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A quadmesh without support
    // of vertex colors.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    // Create an empty fix-sized quadmesh:
    GShadableQuadmesh(int pX, int pY): GTopology2D(pX, pY) {}

    // Create a fix-sized quadmesh and setup the geometry
    // from a x*y sized vertex list:
    GShadableQuadmesh(int pX, int pY, const GVertexAggregate& pVertices):
    GTopology2D(pX, pY, pVertices) {}

    virtual void shade(void);
    // IMPLEMENTED PLATFORM DEPENDENT
};

class GShadablePolyline: public GShadable, virtual public GTopology1D {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A polyline without support
    // of vertex colors.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    // Create an empty polyline:
    GShadablePolyline(void) {}
    // Create a polyline from a const list of vertices:
    GShadablePolyline(const GVertexAggregate& pVertices): GTopology1D(pVertices) {}

    virtual void shade(void);
    // IMPLEMENTED PLATFORM DEPENDENT
};

class GShadablePolygon: public GShadable, virtual public GPolygon {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A planar polygon without support of vertex colors. 
    // 
    // INVARIANT: The polygon must be planar.
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    // Create an empty polygon:
    GShadablePolygon(void) {}
    // Create a polygon from a const list of vertices:
    GShadablePolygon(const GVertexAggregate& pVertices): GPolygon(pVertices)  {}
    virtual void shade(void);
    // IMPLEMENTED PLATFORM DEPENDENT
};

class GShadablePolyhedron: public GShadable, virtual public GTopologyPh {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A shadable polyhedron without supporting vertex colors.
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    // Empty construction: 
    GShadablePolyhedron(void) {}

    // Create it from constant lists:
    GShadablePolyhedron(const GVertexAggregate& pVList, const GFacetAggregate& pFList):
    GTopologyPh(pVList, pFList) {}

    virtual void shade(void);
    // IMPLEMENTED PLATFORM DEPENDENT
};

class GShadeLight {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A class defining the 
    // protocol for "shading" light sources.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    virtual void shade(void) = 0;
};

extern GDLList<GShadeLight>* gShadeLights;

class GShadeAmbientLight: public GShadeLight, virtual public GLight {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A shading compatible ambient light.
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    virtual void shade(void);
    // IMPLEMENTED PLATFORM DEPENDENT
};

class GShadePointLight: public GShadeLight, virtual public GPointLight {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A shading compatible point light.
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    virtual void shade(void);
    // IMPLEMENTED PLATFORM DEPENDENT
};

template<class P>
class GShadingIterator: public GIterator<P> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An iterator that calls the shading
    // methods of primitives or light sources.
    // To be used in shader (sub)class(es).
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: 
    // 
    // CONSTANTS: 
    virtual bool exec(P& p) { 
	p.shade(); 
	return(false);
    }
};

template<class P, class R>
class GShader: public GGRenderer<class P, class R> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An abstract shader to be parametrized
    // with scene aggregate and output realization.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    // get a shadable by tracing the ray:
    virtual P* getRenderable(const GRay3&) = 0;
  protected:
    GShadingIterator<P> mIter;
};

template<class P, class R>
class GLookatShader: public GShader<class P, class R>, public GLookatCamera {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A shader template basing on OpenGLs
    // lookat camera model to be parametrized
    // with scene aggregate and output realization.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    virtual void initRealization(R& r) {
	GGRenderer<P, R>::initRealization(r);
	r.clear(GColorRGB(0.0, 0.0, 0.0));
    }
    virtual void rendering(const GAggregate<P>& a) {
	GGRenderer<P, R>::rendering(a);
	gShaderSetLookatMatrix(*this, *mRealization);
	a.doWithElements(mIter);
	gShaderUnsetMatrix(*this, *mRealization);
    }
    virtual GRay3 inverseTransform(const GPoint2& pPoint, GFloat pAspect = 1.0) const {
	return(GLookatCamera::inverseTransform(pPoint, pAspect));
    }
    virtual P* getRenderable(const GRay3&);
};

template<class P, class R>
inline P* GLookatShader<P, R>::getRenderable(const GRay3& pRay) {
    GGetNearestIterator<P> tIter(pRay);
    assert(mAggregate);
    mAggregate->doWithElements(tIter);
    return(tIter.getRenderable());
}

class GPixmapDisplay;

// Two utility functions to implement the shader functionality:

void gShaderSetLookatMatrix(const GLookatCamera&, const GPixmapDisplay&);
// IMPLEMENTED PLATFORM DEPENDENT
void gShaderUnsetMatrix(const GLookatCamera&, const GPixmapDisplay&);
// IMPLEMENTED PLATFORM DEPENDENT

#endif
