#ifndef G_SOLID_TEXTURE_H
#define G_SOLID_TEXTURE_H

#include "gdata.h"
#include "util/gcolor.h"
#include "util/gmatrix.h"
#include "util/gvector.h"

extern const int gMaxNoise;

typedef enum GSolidTextureSampleMode {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Specifies how to combine
    // texture w/ surface properties.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
    G_NO_SAMPLE = 0, 
    G_REPLACE_SAMPLE, // replace surface components {ambient/
    // diffuse or specular or etc.} of the primitive
    G_ADD_SAMPLE, // add texture to the surface of the primitive
    G_MUL_SAMPLE // multiply texture with the surface of the primitive
};

typedef enum GSolidTextureNormalMode {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Specifies how to combine
    // texture w/ primitive's normal.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
    G_NO_NORMAL = 0, 
    G_REPLACE_NORMAL, // replace the normal of the primitive
    G_ADD_NORMAL // add textures normal to the normal of the primitive
};

class GSolidTexture: public GGraphicalData {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Abstract base class of
    // procedural textures useable for solids.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  protected:
    GSolidTexture(void);
  public:
    virtual GColorRGB getSample(const GPoint3&) const {
	GColorRGB tColor(gBlack);
	return(tColor);
    }
    virtual GVector3 getNormal(const GVector3&) const {
	GVector3 tVector(gOrigin3);
	return(tVector);
    }

    GSolidTextureSampleMode getSampleMode(void) { 
	return(smode); 
    }
    GSolidTextureNormalMode getNormalMode(void) {
	return(nmode); 
    }

    void setMatrix(const GMatrix4&m) {
	mMatrix = m;
    }
    const GMatrix4& get_matrix(void) const {
	return(mMatrix);
    }
  protected:
    static int ***noiseTable;

    GMatrix4 mMatrix;

    GSolidTextureSampleMode smode;
    GSolidTextureNormalMode nmode;

    GFloat getNoise(const GPoint3&) const;
    GColorRGB getRGBNoise(const GPoint3&, GFloat) const;
    GFloat getTurbulence(const GPoint3&, int) const;
};


extern int base(GFloat); 
extern GFloat dsqr(GFloat);
extern GFloat dmod(GFloat, GFloat);
extern GFloat frac(GFloat d);

class GPolyBlobTexture: public GSolidTexture {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: This texture simulates poly blobs.
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GPolyBlobTexture(void) {}
    virtual GColorRGB getSample(const GPoint3 &) const;
};

class GNoiseTexture: public GSolidTexture {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: This texture simulates noise. 
    // The color of the noise can be changed.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    //##### the Tcl/C++ methods:   
    GNoiseTexture(const GColorRGB& _col): col(_col) {}

    void setColor(const GColorRGB& _col) {
	col = _col;
    }
    const GColorRGB& getColor(void) const {
	return(col); 
    }

    virtual GColorRGB getSample(const GPoint3 &) const;
  public:
    GColorRGB col;
};

class GUniColorTexture: public GSolidTexture {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: The texture consists homogeneously
    // of only one color. Objects of this class can be used in other textures.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GUniColorTexture(const GColorRGB& _col): col(_col) {}
    void setColor(const GColorRGB& _col) {
	col = _col; 
    }
    const GColorRGB& getColor(void) const {
	return(col); 
    }
    virtual GColorRGB getSample(const GPoint3 &) const { 
	GColorRGB c(col); 
	return(c); 
    }
  private:
    GColorRGB col;
};

class GWoodTexture: public GSolidTexture {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: This texture simulates 
    // wood consisting of a bright and a dark surface.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GWoodTexture(GSolidTexture* t0, GSolidTexture* t1):
    xlight(0), tex0(t0), tex1(t1) {
	tex0->ref(); tex1->ref();
    }
    virtual ~GWoodTexture(void) {
	tex0->deref();
	tex1->deref(); 
    }

    void setBright(const int i) {
	xlight = i; 
    }
    int getBright(void) const {
	return(xlight); 
    }

    virtual GColorRGB getSample(const GPoint3 &) const;
  private:
    int xlight;
    GSolidTexture* tex0, * tex1;
};

class GCheckerTexture: public GSolidTexture {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Objects of this class create
    // a chessboard-like 3D texture consisting of two other textures.
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GCheckerTexture(GSolidTexture* t0, GSolidTexture* t1):
    tex0(t0), tex1(t1) {
	tex0->ref(); tex1->ref();
    }
    virtual ~GCheckerTexture(void) { 
	tex0->deref(); tex1->deref(); 
    }
    virtual GColorRGB getSample(const GPoint3 &) const;
  private:
    GSolidTexture* tex0, * tex1;
};

class GGridTexture: public GSolidTexture {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: This texture generates a grid consisting of two further textures.
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GGridTexture(GSolidTexture* t0, GSolidTexture* t1):
    xwidth(0.1), tex0(t0), tex1(t1) {
	tex0->ref(); tex1->ref();
    }
    virtual ~GGridTexture(void) {
	tex0->deref(); tex1->deref(); 
    }

    virtual GColorRGB getSample(const GPoint3 &) const;

    void setWidth(const GFloat d) { 
	xwidth = d; 
	if (xwidth < 0) xwidth = 0; 
	if (xwidth > 1) xwidth = 1; 
    }
    GFloat getWidth(void) const {
	return(xwidth); 
    }

  private:
    GFloat xwidth;
    GSolidTexture* tex0, * tex1;
};

class GTurbulenceTexture: public GSolidTexture {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: This texture generates a turbulence
    // from two further surfaces.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GTurbulenceTexture(GSolidTexture* t0, GSolidTexture* t1):
    tex0(t0), tex1(t1), xiter(10), xlevel(0.5) {
	tex0->ref(); tex1->ref();
    }
    virtual ~GTurbulenceTexture(void) {
	tex0->deref(); tex1->deref(); 
    }
    
    virtual GColorRGB getSample(const GPoint3 &) const;

    void setIter(const int i) {
	xiter = i;
	if (xiter < 0) xiter = 0;
    }
    int getIter(void) const {
	return(xiter); 
    }

    void setLevel(const GFloat d) {
	xlevel = d; 
    }
    GFloat getLevel(void) const {
	return(xlevel); 
    }
  private:
    GSolidTexture* tex0, * tex1;
    int xiter;
    GFloat xlevel;
};

class GInterpolationTexture: public GSolidTexture {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: This texture interpolates two further 
    // surfaces in a variable number of steps.
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GInterpolationTexture(GSolidTexture* t0, GSolidTexture* t1):
    tex0(t0), tex1(t1), xsteps(2) {
	tex0->ref(); tex1->ref(); 
    }
    virtual ~GInterpolationTexture(void) {
	tex0->deref(); tex1->deref(); 
    }

    virtual GColorRGB getSample(const GPoint3&) const;

    void setSteps(const int i) { 
	xsteps = i; if (xsteps < 0) xsteps = 0;
    }
    int getSteps(void) const { 
	return(xsteps); 
    }
  private:
    GSolidTexture* tex0, * tex1;
    int xsteps;
};

class GTruncTexture: public GSolidTexture {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: This texture truncs an another continuous texture.
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GTruncTexture(GSolidTexture* _tex): tex(_tex) {
	xsteps = 2; tex->ref();
    }
    virtual ~GTruncTexture(void) { 
	tex->deref(); 
    }

    virtual GColorRGB getSample(const GPoint3 &) const;

    void setSteps(const int i) { 
	xsteps = i; 
	if (xsteps < 0) xsteps = 0;
    }
    int getSteps(void) const {
	return(xsteps); 
    }
  private:
    GSolidTexture* tex;
    int xsteps;
};

class GMarbleTexture: public GSolidTexture {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: This texture simulates marple by linear 
    // interpolating between two other textures. 
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS:  
  public:
    GMarbleTexture(GSolidTexture* _tex0, GSolidTexture* _tex1):
    tex0(_tex0),  tex1(_tex1) {
	tex0->ref(); tex1->ref();
	w[0] = 5; w[1] = 10; w[2] = 15; w[3] = 20;
    }
    virtual ~GMarbleTexture(void) { 
	tex0->deref(); tex1->deref(); 
    }
    virtual GColorRGB getSample(const GPoint3&) const;

    void setWidths(const GFloat _w[4]) {
	w[0] = _w[0]; w[1] = _w[1]; w[2] = _w[2]; w[3] = _w[3];
    }
    GFloat* getWidths(void) const { 
	return((GFloat*)w); 
    };
  private:
    GSolidTexture* tex0, * tex1;
    GFloat w[4];
};

class GXLinearMarbleTexture: public GSolidTexture {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: This texture simulates marple by 
    // linear interpolating between four other textures.
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GXLinearMarbleTexture(GSolidTexture* _tex0, GSolidTexture* _tex1, 
			  GSolidTexture* _tex2, GSolidTexture* _tex3):
    tex0(_tex0), tex1(_tex1), tex2(_tex2), tex3(_tex3) {
	tex0->ref(); tex1->ref(); tex2->ref(); tex3->ref();
	w[0] = 5; w[1] = 10; w[2] = 15; w[3] = 20; w[4] = 25; w[5] = 30; w[6] = 35; 
    }
    virtual ~GXLinearMarbleTexture(void) { 
	tex0->deref(); tex1->deref(); tex2->deref(); tex3->deref(); 
    }

    virtual GColorRGB getSample(const GPoint3&) const;

    void setWidths(const GFloat _w[7]) {
	for (int i = 0; i < 7; i++) 
	    w[i] = _w[i];
    }
    GFloat* getWidths(void) const {
	return((GFloat*)w); 
    }
  private:
    GSolidTexture* tex0, * tex1, * tex2, * tex3;
    GFloat w[7];
};

class GXCubicMarbleTexture: public GSolidTexture {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: This texture simulates marple 
    // by cubic interpolating between four other textures.
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GXCubicMarbleTexture(GSolidTexture* _tex0, GSolidTexture* _tex1, 
			 GSolidTexture* _tex2, GSolidTexture* _tex3) :
    tex0(_tex0), tex1(_tex1), tex2(_tex2), tex3(_tex3) {
	tex0->ref(); tex1->ref(); tex2->ref(); tex3->ref();
	w[0] = 5; w[1] = 10; w[2] = 15; w[3] = 20; w[4] = 25;
    }
    virtual ~GXCubicMarbleTexture(void) { 
	tex0->deref(); tex1->deref(); tex2->deref(); tex3->deref(); 
    }

    virtual GColorRGB getSample( const GPoint3 &) const;

    void setWidths(const GFloat _w[5]) { 
	for (int i = 0; i < 5; i++) 
	    w[i] = _w[i]; 
    }
    GFloat* getWidths(void) const { 
	return((GFloat*)w); 
    }
  private:
    GSolidTexture* tex0, * tex1, * tex2, * tex3;
    GFloat w[5];
};

class GLatticeTexture: public GSolidTexture {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: This demo texture creates a bump mapping 
    // in the x-y plane. The space between the grid is 1, but can 
    // be changed via the texture matrix. This texture is very useful
    // for planar primitives, e.g. a quader or a plane. Surface components
    // will not be changed directly.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 

  public:
    GLatticeTexture(void) { 
	smode = G_NO_SAMPLE;
	nmode = G_REPLACE_NORMAL; 
    }
    virtual GVector3 getNormal(const GPoint3&) const;
};

#endif

