#ifndef G_COLOR_H
#define G_COLOR_H

#include "ggeneric.h"
#include <math.h>

class GColorRGB {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Representation of RGB colors.
    //
    // INVARIANT: Each component must be in the area from 0.0 to 1.0.
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
    friend ostream& operator<<(ostream&, const GColorRGB&);
    // Stream-based output operator.
  public:
    GColorRGB() { mR = mB = mB = 0; }
    // Default constructor; set color to black.
    GColorRGB(GFloat pR, GFloat pG, GFloat pB): mR(pR), mG(pG), mB(pB) {}
    // Constructor with initialization of the components.
    
    // Assignment/modification operators:
    GColorRGB& operator=(const GColorRGB& c) {
	mR = c.mR; mG = c.mG; mB = c.mB; 
	return(*this);
    }
    GColorRGB& operator+=(const GColorRGB& c) {
	mR += c.mR; mG += c.mG; mB += c.mB; 
	return(*this);
    }

    // Set/Get access to red component:
    void setR(GFloat pR) { mR = pR; }
    GFloat getR(void) const { return(mR); }

    // Set/Get access to green component:
    void setG(GFloat pG) { mG = pG; }
    GFloat getG(void) const { return(mG); }

    // Set/Get access to blue component:
    void setB(GFloat pB) { mB = pB; }
    GFloat getB(void) const { return(mB); }

    // Set/Get into/from HSL representation:
    void setHSL(const GFloat[3]);
    void getHSL(GFloat[3]) const;

    // Set color from three float values:
    void setRGB(GFloat pR, GFloat pG, GFloat pB) {
	mR = pR;
	mG = pG;
	mB = pB;
    }

    // Set/Get color from float vector:
    void setRGB(GFloat rgb[3]) {
	mR = rgb[0]; 
	mG = rgb[1];
	mB = rgb[2]; 
    }
    void getRGB(GFloat rgb[3]) const {
	rgb[0] = mR; 
	rgb[1] = mG;
	rgb[2] = mB; 
    }

    // Conversion into an IRIS GL-like long:
    operator unsigned long(void) const;
   
    // Create a new color by multiplication w/ a float:
    GColorRGB operator*(const GFloat x) const {
	return(GColorRGB(mR * x, mG * x, mB* x));
    }

    // Create a new color by multiplication w/ a color:
    GColorRGB operator*(const GColorRGB& x) const {
	return(GColorRGB(mR * x.mR, mG * x.mG, mB * x.mB));
    }

    // Create a new color by addition w/ a color:
    GColorRGB operator+(const GColorRGB& c) const { 
	return(GColorRGB( mR + c.mR, mG + c.mG, mB + c.mB)); 
    }

    // Compare two colors:
    bool operator==(const GColorRGB&) const;

    // Compute the absolute amount:
    GFloat abs(void) const { return(sqrt(mR * mR + mG * mG + mB * mB)); }

    // Clip each wavelength to [0, 1]:
    void correct(void);
  private:
    // Three floats representing the wavelengths for red, green and blue:
    GFloat mR, mG, mB;
};

extern const GColorRGB gBlack;
extern const GColorRGB gWhite;

#endif

