#ifndef G_TEMPLATES_H
#define G_TEMPLATES_H

#include "ggeneric.h"

template<class T>
class GIterator {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Derived types of this class may used to
    // iterate over elements in aggregates. They also may have a 
    // state. However, be carefully with the usage of state-based
    // iterators.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: Passed parameter is a non-const reference
    // in order to allow the calling of non-const methods of T.
    // CONSTANTS:    
  public:
    virtual bool exec(T&) = 0;
    // The abstract iterator functionality.
    // Return true to stop calling this function
    // for further elements.
};

template<class T>
class GOutStreamIterator: public GIterator<T> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Print an aggregate to the
    // specified output stream.
    // The elements are spaced by a white space.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: 
    // 
    // CONSTANTS: 
  public:
    GOutStreamIterator(ostream& pStream): mStream(pStream) {}
    virtual bool exec(T& t) {
	mStream << t << ' ';
	return(false);
    }
  private:
    ostream& mStream;
};

template<class T>
class GPrintIterator: public GIterator<T> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An iterator for demonstration purposes.
    // The functionality is printing an aggregate to the output stream.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: The iterator counts the number of printed
    // elements.
    // CONSTANTS: 
  public:
    GPrintIterator(void): i(0) {}
    virtual bool exec(T& t) {
	if (!i) cout << "printing an aggreate:\n";
	cout << '\t' << i++ << ": " << t << '\n';
	return(false);
    }
    void reset(void) { i = 0; }
  private:
    int i;
};

template<class T>
class GAggregate {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Abstract base class for various aggregate
    // types with list-like topology.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GAggregate(void): mNumber(0) {}

    // Set/Get element by an index:
    virtual void setElement(int, T&) = 0;
    virtual T& getElement(int) const = 0;

    T& operator[](int pIndex) const {
	return(getElement(pIndex));
    }

    virtual void doWithElements(GIterator<T>& ITER) const = 0;
    // Iterate over aggregate elements.
    // ITER is non-const cos the exec function
    // could change its state

    int getNumber(void) const { return(mNumber); }
    // Supply number of elements.
  protected:
    void incNumber(void) { mNumber++; }
    // Increment number after addition of an element.
    void decNumber(void) { mNumber--; }
    // Decrement number after deletion of an element.
    void setNumber(int pNumber) { mNumber = pNumber; }
    // Set number explicitely.
  private:
    int mNumber;
};

template<class T>
class GConstantVector: public GAggregate<T> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A ConstantVector holds a fix number of elements.
    // Elements will be copied by value. All elements will be deleted in
    // destructor.
    // INVARIANT: Number of elements is fixed in constructor.
    // 
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GConstantVector(int pNumber) {
	assert(pNumber > 0);
	setNumber(pNumber);
	mVector = new T[getNumber()];
    }
    virtual ~GConstantVector(void) {
	delete [] mVector;
    }
    virtual void setElement(int pIndex, T& pElement) {
	mVector[pIndex] = pElement;
    }
    virtual T& getElement(int pIndex) const {
	return(mVector[pIndex]);
    }
    virtual void doWithElements(GIterator<T>& func) const {
	T* tVector = mVector;
	for (int i = 0; i < getNumber(); i++) {
	    if (func.exec(*tVector++)) break;
	} 
    }
  private:
    T* mVector;
};

template<class T>
class GDLListElement {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An element for DL lists consisting 
    // of prev and next pointer and an element pointer.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: 
    //
    // CONSTANTS: 
  public:
    GDLListElement(T* pEntry): mEntry(pEntry) {}
    GDLListElement<T>* prev;
    GDLListElement<T>* next;

    // Set/get access:
    void setElement(T& pEntry) {
	delete mEntry;
	mEntry = &pEntry; 
    } 
    T& getElement(void) const { return(*mEntry); }
    T* getPointer(void) const { return(mEntry); }
  private:
    T* mEntry;
};

template<class T>
class GDLList: public GAggregate<T> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A generic DL list holding just 
    // references, not values. 
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: In constructor an automatic deletion
    // of the elements may be requested. Default: off
    //
    // CONSTANTS:     
  public:
    GDLList(bool pClear = false): root(0), last(0), mClear(pClear) {}
    virtual ~GDLList(void);
    void append(T* pElement) { appendAtEnd(pElement); }
    void appendAtBegin(T*);
    void appendAtEnd(T*);
    void remove(T* pElement){ removeFromBegin(pElement); }
    void removeFromBegin(T*);
    void removeFromEnd(T*);

    virtual void setElement(int, T&);
    virtual T& getElement(int) const;
    
    void deleteElement(int);

    virtual void doWithElements(GIterator<T>&) const;
  private:
    GDLListElement<T>* root;
    GDLListElement<T>* last;
    bool mClear;
};

template<class T>
inline GDLList<T>::~GDLList(void) {
    GDLListElement<T>* tmp = root;
    GDLListElement<T>* x;
    if (tmp) do {
	x = tmp;
	if (mClear) delete tmp->getPointer();
	tmp = tmp->next;
	delete x;
    }
    while(tmp);
}

template<class T>
inline void GDLList<T>::appendAtBegin(T* pElement) {
    if (!root) {
	root = new GDLListElement<T>(pElement); 
	root->prev = 0;
	root->next = 0;
	last = root;
    }
    else {
	GDLListElement<T>* tmp = new GDLListElement<T>(pElement); 
	tmp->prev = 0;
	tmp->next = root;
	root->prev = tmp;
	root = tmp;
    }
    incNumber();
}

template<class T>
inline void GDLList<T>::appendAtEnd(T* pElement) {
    if (!last) {
	last = new GDLListElement<T>(pElement); 
	last->prev = 0;
	last->next = 0;
	root = last;
    }
    else {
	GDLListElement<T>* tmp = new GDLListElement<T>(pElement); 
	tmp->prev = last;
	tmp->next = 0;
	last->next = tmp;
	last = tmp;
    }
    incNumber();
}

template<class T>
inline void GDLList<T>::removeFromBegin(T* pElement) { 
    GDLListElement<T>* tmp = root;
    if (!tmp) return;
    do if (&tmp->getElement() == pElement) {
	if (tmp->prev) tmp->prev->next = tmp->next;
	else root = tmp->next;
	if (tmp->next) tmp->next->prev = tmp->prev;
	else last = tmp->prev;
	if (mClear) delete tmp->getPointer();
	delete tmp;
	return;
    }
    while (tmp = tmp->next);
}

template<class T>
inline void GDLList<T>::removeFromEnd(T* pElement) { 
    GDLListElement<T>* tmp = last;
    if (!tmp) return;
    do if (&tmp->getElement() == pElement) {
	if (tmp->prev) tmp->prev->next = tmp->next;
	else root = tmp->next;
	if (tmp->next) tmp->next->prev = tmp->prev;
	else last = tmp->prev;
	if (mClear) delete tmp->getPointer();
	delete tmp;
	return;
    }
    while (tmp = tmp->prev);
}

template<class T>
inline void GDLList<T>::setElement(int pIndex, T& pElement) {
    assert(pIndex >= 0);
    if (pIndex >= getNumber()) append(&pElement);
    else {
	GDLListElement<T>* tmp = root;
	for (int i = 0; i < getNumber(); i++) {
	    assert(tmp); 
	    if (i == pIndex) {
		tmp->setElement(pElement);
		break;
	    }
	    else tmp = tmp->next;
	}
    }
}

template<class T>
inline T& GDLList<T>::getElement(int pIndex) const {
    assert((0 <= pIndex) && (pIndex < getNumber()));
    GDLListElement<T>* tmp = root;
    for (int i = 0; i < getNumber(); i++) {
	assert(tmp); 
	if (i == pIndex) return(tmp->getElement());
	else tmp = tmp->next;
    }
    // never reached cos first assertion:
    return(tmp->getElement());
}

template<class T>
inline void GDLList<T>::deleteElement(int pIndex) { 
    assert((0 <= pIndex) && (pIndex < getNumber()));
    GDLListElement<T>* tmp = root;
    for (int i = 0; i < getNumber(); i++) {
	assert(tmp); 
	if (i == pIndex) {
	    if (tmp->prev) tmp->prev->next = tmp->next;
	    else root = tmp->next;
	    if (tmp->next) tmp->next->prev = tmp->prev;
	    else last = tmp->prev;
	    if (mClear) delete tmp->getPointer();
	    delete tmp;
	    return;
	}
	else tmp = tmp->next;
    }
}

template<class T>
inline void GDLList<T>::doWithElements(GIterator<T>& func) const {
    GDLListElement<T>* tmp = root;
    if (!tmp) return;
    do {
	T& tElement = tmp->getElement();
	tmp = tmp->next;
	if (func.exec(tElement)) break;
    } while(tmp);
}

typedef int GCacheValue;
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An integer-based cache value.
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 

template<class T>
class GCacheTableEntry {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Encapsulation of a cache value
    // and a related DL list.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GCacheTableEntry(GCacheValue pValue): mValue(pValue), mList(true) {}
    GCacheValue getCacheValue(void) const {
	return(mValue);
    }
    GDLList<T>* getCacheList(void) {
	return(&mList);
    }
  private:
    GCacheValue mValue;
    GDLList<T> mList;
};

template <class T>
class GFindCacheListIterator: public GIterator<T> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Find the DL list related to the 
    // initial cache value.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GFindCacheListIterator(GCacheValue pValue): mValue(pValue), mList(0) {}
    virtual bool exec(T& t) {
	if (mValue == t.getCacheValue()) {
	    mList = &t;
	    return(true);
	}
	else return(false);
    }
    T* getList(void) const {
	return(mList);
    }
  private:
    GCacheValue mValue;
    T* mList;
};

template<class T>
class GAppendAggregateIterator: public GIterator<T> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Append all or a specified number of 
    // elements on an existing aggregate,
    // IMPLEMENTATION ISSUES: Use setNumber() to apply one and the same iterator
    // object to different aggregates in case you want to do a 
    // limited appending.
    // CONSTANTS: 
    // 
  public:
    GAppendAggregateIterator(GDLList<T>& pList, bool pIndividualize = false): mList(pList), mIndividualize(pIndividualize), mNumber(0) {}
    virtual bool exec(T& t) {
	if (mIndividualize) mList.append(new T(t));
	else mList.append(&t);
	if (mNumber) {
	    mCounter++;
	    if (mNumber == mCounter) return(true);
	}
	return(false);
    }
    void setNumber(int pNumber) {
	assert(pNumber > 0);
	mNumber = pNumber;
	mCounter = 0;
    }
  private:
    GDLList<T>& mList;
    bool mIndividualize;
    int mNumber;
    int mCounter;
};

template<class T>
class GCopyAggregateIterator: public GIterator<T> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Copy all aggregate elements
    // into a classic C-like array.
    // IMPLEMENTATION ISSUES: Use setNumber() to apply one and the same iterator
    // object to different aggregates in case you want to do a 
    // limited copying.
    // CONSTANTS: 
    // 
  public:
    GCopyAggregateIterator(T* pVector): mVector(pVector), mNumber(0) {}
    virtual bool exec(T& t) {
	*mVector++ = t;
	return(false);
    } 
    void setNumber(int pNumber) {
	assert(pNumber > 0);
	mNumber = pNumber;
	mCounter = 0;
    }
  private:
    T* mVector;
    int mNumber;
    int mCounter;
};

template<class T>
class GCopyAggregateIterator2: public GIterator<T> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Copy all aggregate elements
    // into an another aggregate.
    // IMPLEMENTATION ISSUES: 
    // 
    // CONSTANTS: 
  public:
    GCopyAggregateIterator2(GAggregate<T>& pAggr): mAggr(pAggr), mCounter(0) {}
    virtual bool exec(T& t) {
	mAggr.setElement(mCounter++, t);
	return(false);
    } 
  private:
    GAggregate<T>& mAggr;
    int mCounter;
};

template<class T>
class GCompareAggregateIterator: public GIterator<T> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Compare two aggregate by
    // checking the aquivalency of the elements.
    // IMPLEMENTATION ISSUES: 
    // 
    // CONSTANTS: 
  public:
    GCompareAggregateIterator(const GAggregate<T>& pAggr): mAggr(pAggr), mCounter(0), mEqual(true) {}
    virtual bool exec(T& t) {
	if (t == mAggr.getElement(mCounter++)) return(false);
	else {
	    mEqual = false;
	    return(true);
	}
    } 
    bool wasEqual(void) const { return(mEqual); }
  private:
    const GAggregate<T>& mAggr;
    int mCounter;
    bool mEqual;
};

template <class T>
class GFindByNameIterator: public GIterator<T> {
    // CONCEPT: Furniture graphics kernel
    //
    // ABSTRACT PURPOSE: Look for object by
    // specified name.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GFindByNameIterator(const char* pName): mName(pName), mT(0) {}
    virtual bool exec(T& t) {
	if (t.getName() == mName) {
	    mT = &t;
	    return(true);
	}
	else return(false);
    }
    T* get(void) { return(mT); }
  private:
    const char* mName;
    T* mT;
};

template <class T>
class GFindElementIterator: public GIterator<T> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Find an element
    // by checking address identity.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GFindElementIterator(const T* pElement): mElement(pElement), mFound(false) {}
    virtual bool exec(T& t) {
	if (&t == mElement) {
	    mFound = true;
	    return(true);
	}
	else return(false);
    }
    bool found(void) const { return(mFound); }
  private:
    const T* mElement;
    bool mFound;
};

#endif

