#ifndef G_VECTOR_H
#define G_VECTOR_H

#include <math.h>
#include <stdio.h>
#include "ggeneric.h"

class GVector2 {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A type consisting 
    // of two scalars.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
    friend ostream& operator<<(ostream&, const GVector2&);
  public:
    GVector2(void) { mX = mY = 0; }
    GVector2(GFloat pVal[2]): mX(pVal[0]), mY(pVal[1]) {}
    GVector2(GFloat pX, GFloat pY): mX(pX), mY(pY) {} 
    GVector2(GFloat pF): mX(pF), mY(pF) {} 

    void setX(GFloat pX) { mX = pX; }
    GFloat getX(void) const { return mX; }

    void setY(GFloat pY) { mY = pY; }
    GFloat getY(void) const { return mY; }

    bool operator==(const GVector2&) const;
  private:
    GFloat mX, mY;
};

typedef GVector2 GPoint2;

class GVector3 {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A basic data type
    // for 3D graphics: the 3D vector or point.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
    friend class GMatrix4;
    friend ostream& operator<<(ostream&, const GVector3&);
    friend GVector3 gCalcVUP(const GVector3&, const GVector3&, GFloat);
  public:
    GVector3(void) { mX = mY = mZ = 0; }
    GVector3(GFloat pVal[3]): mX(pVal[0]), mY(pVal[1]), mZ(pVal[2]) {}
    GVector3(GFloat pX, GFloat pY, GFloat pZ): mX(pX), mY(pY), mZ(pZ) {} 
    GVector3(GFloat pF): mX(pF), mY(pF), mZ(pF) {} 

    void setX(GFloat pX) { mX = pX; }
    GFloat getX(void) const { return mX; }

    void setY(GFloat pY) { mY = pY; }
    GFloat getY(void) const { return mY; }

    void setZ(GFloat pZ) { mZ = pZ; }
    GFloat getZ(void) const { return mZ; }


    bool operator==(const GVector3&) const;

    GVector3 operator+(const GVector3& v) const { 
	GVector3 c(mX + v.mX, mY + v.mY, mZ + v.mZ);
	return(c);
    }
    GVector3 operator-(const GVector3& v) const {
	GVector3 c(mX - v.mX, mY - v.mY, mZ - v.mZ);
	return(c); 
    }

    GVector3 operator*(const GFloat d) const {
	GVector3 c(mX * d, mY * d, mZ * d);
	return(c);  
    }

    double abs(void) const {
	return(sqrt(mX * mX + mY * mY + mZ * mZ));
    }

    GVector3 scale(GFloat f) const {	
	return(*this * f);
    }

    double dot(const GVector3& v) const {
	return(mX * v.mX + mY * v.mY + mZ * v.mZ);    
    }

    GVector3 cross(const GVector3& v) const {
	GVector3 c(mY * v.mZ - mZ * v.mY, 
		   mZ * v.mX - mX * v.mZ,
		   mX * v.mY - mY * v.mX);
	return(c);   
    }

    GVector3 eval(const GVector3& v) const {
	GVector3 c(mX * v.mX, mY * v.mY, mZ * v.mZ );
	return(c);  
    }

    GVector3 unitize(void) const {
	double r = abs();
	assert(r != 0.0);
	GVector3 c(mX / r, mY / r, mZ / r);
	return(c);
    }

    // get into OpenGL/IRIS GL like float array:
    void get(float f[3]) const { 
	f[0] = mX; f[1] = mY; f[2] = mZ; 
    }

    GVector3 negate(void) {
	GVector3 c(-mX, -mY, -mZ );
	return(c); 
    }

    GFloat& operator[] (int inx) {
	if (inx == 0) return(mX);
	else if (inx == 1) return(mY);
	else return(mZ);
    }

    const GFloat& operator[] (int inx) const {
	if (inx == 0) return(mX);
	else if (inx == 1) return(mY);
	else return(mZ);
    }

    GVector3& operator-=(const GVector3& A) {
	mX -= A.mX;  
	mY -= A.mY;  
	mZ -= A.mZ;
	return(self);
    }

    GVector3& operator*=(GFloat A) {
	mX *= A;  
	mY *= A;  
	mZ *= A;
	return(self);
    }

    GVector3& operator/=(GFloat A) {
	mX /= A;  
	mY /= A; 
	mZ /= A;
	return(self);
    }

    GVector3& operator*=(const GVector3& A) {
	mX *= A.mX; 
	mY *= A.mY; 
	mZ *= A.mZ;
	return(self);
    }

friend GVector3 operator+(const GVector3& A, const GVector3& B) {
    return(GVector3(A.mX + B.mX, A.mY + B.mY, A.mZ + B.mZ));
}

friend GVector3 operator-(const GVector3& A, const GVector3& B) {
    return(GVector3(A.mX - B.mX, A.mY - B.mY, A.mZ - B.mZ));
}

friend GVector3 operator*(const GVector3& A, const GVector3& B) {
    return(GVector3(A.mX * B.mX, A.mY * B.mY, A.mZ * B.mZ));
}

friend GVector3 operator*(const GVector3& A, GFloat B) {
    return(GVector3(A.mX * B, A.mY * B, A.mZ * B));
}

friend GVector3 operator*(GFloat A, const GVector3& B) {
    return(B * A);
}

friend GVector3 operator/(const GVector3& A, const GVector3& B) {
    return(GVector3(A.mX / B.mX, A.mY / B.mY, A.mZ / B.mZ));
}

friend GVector3 operator/(const GVector3& A, GFloat B) {
    return(GVector3(A.mX / B, A.mY / B, A.mZ / B));
}

friend GVector3 getMin(const GVector3& a, const GVector3& b) {
    // Return a vector containing the minimal components of the 
    // passed vectors.
    GVector3 tVec;
    tVec.mX = gmin(a.mX, b.mX);
    tVec.mY = gmin(a.mY, b.mY);
    tVec.mZ = gmin(a.mZ, b.mZ);
    return(tVec);
}

friend GVector3 getMax(const GVector3& a, const GVector3& b) {
    // Return a vector containing the maximal components of the 
    // passed vectors.
    GVector3 tVec;
    tVec.mX = gmax(a.mX, b.mX);
    tVec.mY = gmax(a.mY, b.mY);
    tVec.mZ = gmax(a.mZ, b.mZ);
    return(tVec);
}

    void setMin(const GVector3& a_) {
	// METHOD: Apply a minimum vector to this vector.
	if (a_.mX < mX) mX = a_.mX;
	if (a_.mY < mY) mY = a_.mY;
	if (a_.mZ < mZ) mZ = a_.mZ;
    }

    void setMax(const GVector3& a_) {
	// METHOD: Apply a maximum vector to this vector.
	if(a_.mX > mX) mX = a_.mX;
	if(a_.mY > mY) mY = a_.mY;
	if(a_.mZ > mZ) mZ = a_.mZ;
    }
    
    void project(GFloat& px, GFloat& py, int axis) const{
	// METHOD: Project the point to a plane.
	switch (axis) {
	  case 0:
	    px = mY;
	    py = mZ;
	    break;
	  case 1:
	    px = mX;
	    py = mZ;
	    break;
	  default:
	    px = mX;
	    py = mY;
	    break;
	}
    }
  private:
    GFloat mX, mY, mZ;
};

typedef GVector3 GPoint3;
typedef GVector3 GBounds;
typedef GVector3 GNormal;

extern const GPoint3 gOrigin3;

class GRay3 {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A simple representation
    // of a linear ray.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: The length is optional.
    // Its interpretation depends on the scaling of the
    // direction vector.
    // CONSTANTS: 
    friend ostream& operator<<(ostream&, const GRay3&);

 public:
    // ray types 
    enum  {
      R_UNKNOW  = 0x00,
	R_SHADOW  = 0x01,
	R_REFLECT = 0x02,
	R_REFRACT = 0x04,
	R_EYE     = 0x08
    };
    
    GRay3(void): mLength(0), mTmin(gEpsilon), mTmax(HUGE), mFlags(R_UNKNOW) {}
    GRay3(const GPoint3& pPoint, const GVector3& pVector, int pFlags=R_UNKNOW)
      : mPoint(pPoint), 
      mVector(pVector), 
      mLength(0), 
      mTmin(gEpsilon), 
      mTmax(HUGE), 
      mFlags(pFlags) {}

    // set/get coordinate:
    void setPoint(const GPoint3& pPoint) { mPoint = pPoint; }
    const GPoint3 getPoint(void) const { return(mPoint); }

    // set/get direction vector:
    void setVector(const GVector3& pVector) { mVector = pVector; }
    const GVector3 getVector(void) const { return(mVector); }

    // set/get length:
    void setLength(GFloat pLength) { mLength = pLength; }
    GFloat getLength(void) const { return(mLength); }

    // compute 3D point for passed distance:
    GPoint3 compute(GFloat t) const {
	return(mPoint + mVector * t);
    }
    bool isValid(GFloat t) const {
	return((t > mTmin) && (t < mTmax));
    }
    bool isValid(GFloat bmin, GFloat bmax) const {
	// sort it:
	if (bmin > bmax) {
	    GFloat t = bmin; 
	    bmin = bmax; 
	    bmax = t;
	}
	return(!((bmax < mTmin) || (bmin > mTmax)));
    }
    
    void setTmax(GFloat pTmax) { mTmax = pTmax; }
    GFloat getTmax(void) const { return(mTmax); }
    
    void setTmin(GFloat pTmin) { mTmin = pTmin; }
    GFloat getTmin(void) const { return(mTmin); }
  
    bool isShadowRay(void) const {
	return(mFlags & R_SHADOW);
    }
    bool isReflectRay(void) const{
	return(mFlags & R_REFLECT);
    }
    bool isRefractRay(void)  const{
	return (mFlags & R_REFRACT);
    }
    bool isEyeRay(void)  const{
	return (mFlags & R_EYE);
    }
   
    // Compute the point w/ distance t to the ray origin 
    GPoint3 extrap(GFloat pT) const {
	return(mPoint + mVector * pT);
    }
 
    // update max. and min. extensions 
    void minmax(GFloat pMin, GFloat pMax) {
	if (pMin > pMax) { 
	    GFloat t = pMin;
	    pMin = pMax;
	    pMax = t;
	}
	
	mTmin = gmax(pMin, mTmin);
	mTmax = gmin(pMax, mTmax);
    }
    
 private:
    GPoint3 mPoint;
    GVector3 mVector;
    GFloat mLength;
    GFloat mTmin, mTmax;
    int mFlags;
};

class GBounds3 {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An orthogonal box
    // in world coordinates.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: 
    // 
    // CONSTANTS: 
    friend ostream& operator<<(ostream&, const GBounds3&);
  public:
    GBounds3(void) {}
    GBounds3(const GVector3& pMin, const GVector3& pMax): mMin(pMin), mMax(pMax) {}

    // set/get minimum:
    void setMin(const GVector3& pMin) { mMin = pMin; }
    const GVector3& getMin(void) const { return(mMin); }

    // set/get maximum:
    void setMax(const GVector3& pMax) { mMax = pMax; }
    const GVector3& getMax(void) const { return(mMax); }

  private:
    GVector3 mMin, mMax;
};

#endif

