#ifndef G_VERTEX_H
#define G_VERTEX_H

#include "util/gvector.h"
#include "util/gcolor.h"
#include "util/gsurface.h"

class GVertex {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Representation of vertices 
    // consisting of 3D geometry and an optional normal
    // vector.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: 
    //
    // CONSTANTS: 
  public:
    // Default initializer:
    GVertex(void): mNormal(0) {}
    // Initialize geometry:
    GVertex(const GVector3& pPoint): mPoint(pPoint), mNormal(0) {}
    // Initialize geometry and normal vector:
    GVertex(const GVector3& pPoint, const GVector3 &pNormal): mPoint(pPoint), mNormal(0) { 
	setNormal(&pNormal);
    }
    // Copy constructor:
    GVertex(const GVertex& pVertex): mNormal(0) {
	setPoint(pVertex.getPoint());
	setNormal(pVertex.getNormal());
    }

    virtual ~GVertex(void) { if (mNormal) delete mNormal; }

    // Assigment operator:
    GVertex& operator=(const GVertex& pVertex) {
	setPoint(pVertex.getPoint());
	setNormal(pVertex.getNormal());
	return(*this);
    }

    // Set/get point:
    void setPoint(const GVector3& pPoint) { mPoint = pPoint; }
    const GVector3& getPoint(void) const { return(mPoint); }

    // Set/get normal vector:
    void setNormal(const GVector3* pNormal) {
	if (!pNormal) {
	    if (mNormal) {
		delete mNormal;
		mNormal = 0;
	    }
	}
	else {
	    if (!mNormal) mNormal = new GVector3;
	    *mNormal = *pNormal;
	}
    }
    const GVector3* getNormal(void) const { return(mNormal); }
  private:
    GVector3 mPoint;
    GVector3* mNormal;
};

ostream& operator<<(ostream&, const GVertex&);

class GVertexWithColorRGB: public GVertex {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A vertex additionally
    // providing an RGB color (optionally).
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    // Default initializer:
    GVertexWithColorRGB(void): mColor(0) {}
    // Initialize geometry:
    GVertexWithColorRGB(const GVector3& pPoint): GVertex(pPoint), mColor(0) {}
    // Initialize geometry and normal vector:
    GVertexWithColorRGB(const GVector3& pPoint, const GVector3 &pNormal): GVertex(pPoint, pNormal), mColor(0) {}
    // Initialize geometry and color:
    GVertexWithColorRGB(const GVector3& pPoint, const GColorRGB& pColor): GVertex(pPoint), mColor(0) {
	setColor(&pColor);
    }
    // Initialize all components:
    GVertexWithColorRGB(const GVector3& pPoint, const GVector3 &pNormal, const GColorRGB& pColor): GVertex(pPoint, pNormal), mColor(0) {
	setColor(&pColor);
    }

    // Copy constructor:
    GVertexWithColorRGB(const GVertexWithColorRGB& pVertex) {
	setPoint(pVertex.getPoint());
	setNormal(pVertex.getNormal());
	setColor(pVertex.getColor());
    }

    virtual ~GVertexWithColorRGB(void) { if (mColor) delete mColor; }

    // Assigment operator:
    GVertexWithColorRGB& operator=(const GVertexWithColorRGB& pVertex) {
	(GVertex&)*this = pVertex;
	setColor(pVertex.getColor());
	return(*this);
    }

    // Set/get color value:
    void setColor(const GColorRGB* pColor) {
	if (!pColor) {
	    if (mColor) {
		delete mColor;
		mColor = 0;
	    }
	}
	else {
	    if (!mColor) mColor = new GColorRGB;
	    *mColor = *pColor;
	}
    }
    const GColorRGB* getColor(void) const { return(mColor); }
  private:
    GColorRGB* mColor;
};

ostream& operator<<(ostream&, const GVertexWithColorRGB&);

class GVertexWithSurface: public GVertex {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A vertex additionally
    // providing a surface (optionally).
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    // Default initializer:
    GVertexWithSurface(void): mSurface(0) {}
    // Initialize geometry:
    GVertexWithSurface(const GVector3& pPoint): GVertex(pPoint), mSurface(0) {}
    // Initialize geometry and normal vector:
    GVertexWithSurface(const GVector3& pPoint, const GVector3 &pNormal): GVertex(pPoint, pNormal), mSurface(0) {}
    // Initialize geometry and surface:
    GVertexWithSurface(const GVector3& pPoint, const GSurface& pSurface): GVertex(pPoint), mSurface(0) {
	setSurface(&pSurface);
    }
    // Initialize all components:
    GVertexWithSurface(const GVector3& pPoint, const GVector3 &pNormal, const GSurface& pSurface): GVertex(pPoint, pNormal), mSurface(0) {
	setSurface(&pSurface);
    }

    // Copy constructor:
    GVertexWithSurface(const GVertexWithSurface& pVertex) {
	setPoint(pVertex.getPoint());
	setNormal(pVertex.getNormal());
	setSurface(pVertex.getSurface());
    }

    virtual ~GVertexWithSurface(void) { if (mSurface) delete mSurface; }

    // Assigment operator:
    GVertexWithSurface& operator=(const GVertexWithSurface& pVertex) {
	(GVertex&)*this = pVertex;
	setSurface(pVertex.getSurface());
	return(*this);
    }

    // Set/get surface pointer:
    void setSurface(const GSurface* pSurface) {
	if (!pSurface) {
	    if (mSurface) {
		delete mSurface;
		mSurface = 0;
	    }
	}
	else {
	    if (!mSurface) mSurface = new GSurface;
	    *mSurface = *pSurface;
	}
    }
    const GSurface* getSurface(void) const { return(mSurface); }
  private:
    GSurface* mSurface;
};

ostream& operator<<(ostream&, const GVertexWithSurface&);

#endif


