//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel. Copying, distribution and     
//  legal info is in the file COPYRIGHT which should be distributed with this 
//  file. If COPYRGHT is not available or for more info please contact: 
//		ekki@prakinf.tu-ilmenau.de
/////////// (C) '95 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include <GL/gl.h>
#include <GL/glu.h>

#include "goutput.h"
#include "render/gshade.h"
#include "render/gpixmap.h"
#include "input/gevent.h"

static int gpLightCnt;

// class GShadable:

GShadable::~GShadable(void) {
    if (getID()) {
	if (glIsList((GLuint)getID())) glDeleteLists((GLuint)getID(), (GLsizei)1);
	gFreeID(getID());
    }
}

void GShadable::setMatrix(const GMatrix4& pMat) {
    GLfloat m[16];
    int ind = 0;
    for (int i = 0; i < 4; i++)
	for (int j = 0; j < 4; j++)
	    m[ind++] = pMat.get(j, i);
    glPushMatrix();
    glMultMatrixf(m);
}

void GShadable::unsetMatrix(void) { 
    glPopMatrix(); 
}

void GShadable::setSurface(const GSurface& pSurface) {
    glColorMaterial(GL_FRONT, GL_AMBIENT);
    glColor3f(pSurface.getAmbient().getR(), pSurface.getAmbient().getG(), pSurface.getAmbient().getB());
    glColorMaterial(GL_FRONT, GL_DIFFUSE);
    glColor3f(pSurface.getDiffuse().getR(), pSurface.getDiffuse().getG(), pSurface.getDiffuse().getB());
    glColorMaterial(GL_FRONT, GL_SPECULAR);
    glColor3f(pSurface.getSpecular().getR(), pSurface.getSpecular().getG(), pSurface.getSpecular().getB());
}

void GShadable::openDisplayList(void) {
    if (!mID) mID = gGetID();
    if (glIsList((GLuint)getID())) glDeleteLists((GLuint)getID(), (GLsizei)1);
    glNewList((GLuint)getID(), GL_COMPILE);
}

void GShadable::closeDisplayList(void) { 
    glEndList(); 
}

void GShadable::callDisplayList(void) {
    if (!getID()) gOutput->fatal("Attempt to call non-existing display list.");
    glCallList((GLuint)getID());
} 

// class GShadableQuadmesh:

void GShadableQuadmesh::shade(void) {
    GVector3 defNormal(0, 0, 1);
    int rows = getX();
    int cols = getY();
    float f[4];

    if (fillstyle) {
	for (int i = 0; i < rows - 1; i++) {
	    glBegin(GL_QUAD_STRIP);
	    for (int j = 0; j < cols; j++) {
		for (int p = 0; p < 2; p++) {
		    const GVertex& vt = getVertex(i + p, j);
		    if (vt.getNormal()) { 
			vt.getNormal()->get(f); 
		    }
		    else defNormal.get(f);
		    glNormal3fv(f);
		    vt.getPoint().get(f);
		    glVertex3fv(f);
		}
	    }   
	    glEnd();
	}
    } else {
	for (int i = 0; i < rows; i++) {
	    glBegin(GL_LINE_STRIP);
	    for(int j = 0; j < cols; j++) {
		const GVertex& vt = getVertex(i, j);  
		vt.getPoint().get(f);
		glVertex3fv(f);
	    }
	    glEnd();
	}
	for (int j = 0; j < cols; j++) {
	    glBegin(GL_LINE_STRIP);
	    for (int i = 0; i < rows; i++) {
		const GVertex& vt = getVertex(i, j);  
		vt.getPoint().get(f);
		glVertex3fv(f);
	    }
	    glEnd();
	}
    }
}

template<class V>
class GShadeTopology1DIterator: public GIterator<V> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Shading iterator for subclasses
    // of 1D topology.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    virtual bool exec(V& v) { 
	GLfloat p[4];
	const GVector3* n = v.getNormal();
	if (n) {
	    n->get(p); 
	    glNormal3fv(p);
	}
	GVector3 po = v.getPoint();
	p[0] = po.getX();
	p[1] = po.getY();
	p[2] = po.getZ();
	glVertex3fv(p);
	return(false);
    }
  private:
    const GVector3* mNormal;
};

void GShadablePolyline::shade(void) {
    glBegin(GL_LINE_STRIP);
    GShadeTopology1DIterator<GVertex> tIter;
    getVertexList().doWithElements(tIter);
    glEnd();
}

void GShadablePolygon::shade(void) {
    if (fillstyle) {
	glBegin(GL_POLYGON); 
	GLfloat n[4];
	getNormal().get(n);
	glNormal3fv(n); 
    }
    else glBegin(GL_LINE_LOOP);
    GShadeTopology1DIterator<GVertex> tIter;
    getVertexList().doWithElements(tIter);
    glEnd();
}

template<class F>
class GShadeTopologyPhIterator: public GIterator<F> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An iterator for shading polyhedron-compatible
    // shadable primitives.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: The iterator is only useable to render
    // exactly one polyhedron.
    // CONSTANTS: 
  public:
    GShadeTopologyPhIterator(const GTopologyPh& pPolyhedron): mPolyhedron(pPolyhedron), mNumber(0) {}
    virtual bool exec(F& f) { 
	GLfloat sFloat[4]; 
    	if (GShadable::fillstyle) glBegin(GL_POLYGON);
   	else glBegin(GL_LINE_LOOP);	
	for (int i = 0; i < f.getNumber(); i++) {
	    const GVertex& v = mPolyhedron.getVertex(mNumber, i);
	    if (v.getNormal()) {
		v.getNormal()->get(sFloat);
		glNormal3fv(sFloat); 
	    }
	    v.getPoint().get(sFloat);
	    glVertex3fv(sFloat);
	}
	glEnd();
	mNumber++;
	return(false);
    }
  private:
    const GTopologyPh& mPolyhedron;
    int mNumber;
};

void GShadablePolyhedron::shade(void) {
    GShadeTopologyPhIterator<GFacet> tIter(*this);
    getFacetList().doWithElements(tIter);
}

void gShaderSetLookatMatrix(const GLookatCamera& r, const GPixmapDisplay& p) {
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glEnable(GL_COLOR_MATERIAL); 
    glEnable(GL_NORMALIZE);  
#ifdef MESA
    glShadeModel(GL_FLAT);
#else
    glShadeModel(GL_SMOOTH);
#endif
    glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);
    glDrawBuffer(GL_BACK);
    glFrontFace(GL_CW);

    glPushMatrix();
    // setup projection:
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    if (r.isPerspectivic()) {
	gluPerspective((GLdouble)r.getZoom(), (GLdouble)p.getWidth()/(GFloat)p.getHeight(), (GLdouble)r.getNear(), (GLdouble)r.getFar());
    }
    else {
	GFloat tWidth = tan((GLdouble)r.getZoom() * M_PI_2 / 90.0);
	GFloat tHeight = tWidth;
	if (p.getWidth() > p.getHeight()) {
	    tWidth *= (GLdouble)p.getWidth()/(GFloat)p.getHeight();
	}
	if (p.getHeight() > p.getWidth()) {
	    tHeight *= (GLdouble)p.getHeight()/(GFloat)p.getWidth();
	}

	glOrtho(-tWidth, tWidth, -tHeight, tHeight, (GLdouble)r.getNear(), (GLdouble)r.getFar());
    }

    GVector3 vup = gCalcVUP(r.getViewpoint(), r.getRefpoint(), r.getTwist() * M_PI / 180.0);
    gluLookAt(r.getViewpoint().getX(), r.getViewpoint().getY(), r.getViewpoint().getZ(),
	      r.getRefpoint().getX(), r.getRefpoint().getY(), r.getRefpoint().getZ(), 
	      -vup.getX(), -vup.getY(), -vup.getZ()
	      );    
    
    // switch in model mode:
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    // reset light sources:
    gpLightCnt = 0;

    for (int i = 0; i < 8; i++) glDisable(GL_LIGHT0 + i);

    // setup the light sources:
    GShadingIterator<GShadeLight> tIter;
    if (gShadeLights) gShadeLights->doWithElements(tIter);
} 

int gGetID(void) {
    static sNumber = 1;
    return(sNumber++);
}

void gFreeID(int) {}

void gShaderUnsetMatrix(const GLookatCamera&, const GPixmapDisplay&) {
    glPopMatrix();
}

void GShadePointLight::shade(void) {
    if (!isOn()) return;

    GLenum tEnum = GL_LIGHT0 + gpLightCnt++;

    GLfloat color[4]; 
    int n = 0; 
    color[n++] = 0.0; 
    color[n++] = 0.0; 
    color[n++] = 0.0; 
    color[n++] = 1.0;
    glLightfv(tEnum, GL_AMBIENT, color);

    n = 0;
    color[n++] = 0.7; 
    color[n++] = 0.7; 
    color[n++] = 0.3; 
    color[n++] = 1.0;
    glLightfv(tEnum, GL_DIFFUSE, color);

    n = 0;
    color[n++] = 0.7;
    color[n++] = 0.7;
    color[n++] = 0.3;
    color[n++] = 1.0;
    glLightfv(tEnum, GL_SPECULAR, color);

    GLfloat geometry[4];
    n = 0;
    geometry[n++] = getPosition().getX();
    geometry[n++] = getPosition().getY();
    geometry[n++] = getPosition().getZ();
    geometry[n++] = 0.0;
    glLightfv(tEnum, GL_POSITION, geometry);

    GLfloat sgl[1];
    n = 0;
    sgl[n++] = 10.0;
    glLightfv(tEnum, GL_SPOT_EXPONENT, sgl); 
    
    n = 0;
    sgl[n++] = 360.0;
    glLightfv(tEnum, GL_SPOT_CUTOFF, sgl); 
    
    glEnable(tEnum);
}

void GShadeAmbientLight::shade(void) {}

