//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel. Copying, distribution and     
//  legal info is in the file COPYRIGHT which should be distributed with this 
//  file. If COPYRGHT is not available or for more info please contact: 
//		ekki@prakinf.tu-ilmenau.de
/////////// (C) '95 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include <GL/gl.h>
#include <GL/glu.h>

#include "goutput.h"
#include "render/gshade.h"
#include "render/gpixmap.h"
#include "input/gevent.h"

static int gpLightCnt;

// class GShadable:

GShadable::~GShadable(void) {
    if (getID()) {
	if (glIsList((GLuint)getID())) glDeleteLists((GLuint)getID(), (GLsizei)1);
	gFreeID(getID());
    }
}

void GShadable::setMatrix(const GMatrix4& pMat) {
    GLfloat m[16];
    int ind = 0;
    for (int i = 0; i < 4; i++)
	for (int j = 0; j < 4; j++)
	    m[ind++] = pMat.get(j, i);
    glPushMatrix();
    glMultMatrixf(m);
}

void GShadable::unsetMatrix(void) { 
    glPopMatrix(); 
}

void GShadable::setSurface(const GSurface&) {}

void GShadable::openDisplayList(void) {
    if (!mID) mID = gGetID();
    if (glIsList((GLuint)getID())) glDeleteLists((GLuint)getID(), (GLsizei)1);
    glNewList((GLuint)getID(), GL_COMPILE);
}

void GShadable::closeDisplayList(void) { 
    glEndList(); 
}

void GShadable::callDisplayList(void) {
    if (!getID()) gOutput->fatal("Attempt to call non-existing display list.");
    glCallList((GLuint)getID());
} 

// class GShadableQuadmesh:

void GShadableQuadmesh::shade(void) {
    int rows = getX();
    int cols = getY();
    float f[4];

    if (fillstyle) {
	for (int i = 0; i < rows - 1; i++) {
	    glBegin(GL_QUAD_STRIP);
	    for (int j = 0; j < cols; j++) {
		for (int p = 0; p < 2; p++) {
		    const GVertex& vt = getVertex(i + p, j);
		    vt.getPoint().get(f);
		}
	    }   
	    glEnd();
	}
    } else {
	for (int i = 0; i < rows; i++) {
	    glBegin(GL_LINE_STRIP);
	    for(int j = 0; j < cols; j++) {
		const GVertex& vt = getVertex(i, j);  
		vt.getPoint().get(f);
		glVertex3fv(f);
	    }
	    glEnd();
	}
	for (int j = 0; j < cols; j++) {
	    glBegin(GL_LINE_STRIP);
	    for (int i = 0; i < rows; i++) {
		const GVertex& vt = getVertex(i, j);  
		vt.getPoint().get(f);
		glVertex3fv(f);
	    }
	    glEnd();
	}
    }
}

template<class V>
class GShadeTopology1DIterator: public GIterator<V> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Shading iterator for subclasses
    // of 1D topology.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    virtual bool exec(V& v) { 
	GVector3 po = v.getPoint();
	GLfloat p[4];
	p[0] = po.getX();
	p[1] = po.getY();
	p[2] = po.getZ();
	glVertex3fv(p);
	return(false);
    }
  private:
    const GVector3* mNormal;
};

void GShadablePolyline::shade(void) {
    glBegin(GL_LINE_STRIP);
    GShadeTopology1DIterator<GVertex> tIter;
    getVertexList().doWithElements(tIter);
    glEnd();
}

void GShadablePolygon::shade(void) {
    glBegin(GL_LINE_LOOP);
    GShadeTopology1DIterator<GVertex> tIter;
    getVertexList().doWithElements(tIter);
    glEnd();
}

template<class F>
class GShadeTopologyPhIterator: public GIterator<F> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An iterator for shading polyhedron-compatible
    // shadable primitives.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: The iterator is only useable to render
    // exactly one polyhedron.
    // CONSTANTS: 
  public:
    GShadeTopologyPhIterator(const GTopologyPh& pPolyhedron): mPolyhedron(pPolyhedron), mNumber(0) {}
    virtual bool exec(F& f) { 
	GLfloat sFloat[4]; 
   	glBegin(GL_LINE_LOOP);	
	for (int i = 0; i < f.getNumber(); i++) {
	    const GVertex& v = mPolyhedron.getVertex(mNumber, i);
	    v.getPoint().get(sFloat);
	    glVertex3fv(sFloat);
	}
	glEnd();
	mNumber++;
	return(false);
    }
  private:
    const GTopologyPh& mPolyhedron;
    int mNumber;
};

void GShadablePolyhedron::shade(void) {
    GShadeTopologyPhIterator<GFacet> tIter(*this);
    getFacetList().doWithElements(tIter);
}

void gShaderSetLookatMatrix(const GLookatCamera& r, const GPixmapDisplay& p) {
    glShadeModel(GL_FLAT);
    glDrawBuffer(GL_BACK);
    
    glPushMatrix();
    // setup projection:
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    if (r.isPerspectivic()) {
	gluPerspective((GLdouble)r.getZoom(), (GLdouble)p.getWidth()/(GFloat)p.getHeight(), (GLdouble)r.getNear(), (GLdouble)r.getFar());
    }
    else {
	GFloat tWidth = tan((GLdouble)r.getZoom() * M_PI_2 / 90.0);
	GFloat tHeight = tWidth;
	if (p.getWidth() > p.getHeight()) {
	    tWidth *= (GLdouble)p.getWidth()/(GFloat)p.getHeight();
	}
	if (p.getHeight() > p.getWidth()) {
	    tHeight *= (GLdouble)p.getHeight()/(GFloat)p.getWidth();
	}

	glOrtho(-tWidth, tWidth, -tHeight, tHeight, (GLdouble)r.getNear(), (GLdouble)r.getFar());
    }

    GVector3 vup = gCalcVUP(r.getViewpoint(), r.getRefpoint(), r.getTwist() * M_PI / 180.0);
    gluLookAt(r.getViewpoint().getX(), r.getViewpoint().getY(), r.getViewpoint().getZ(),
	      r.getRefpoint().getX(), r.getRefpoint().getY(), r.getRefpoint().getZ(), 
	      -vup.getX(), -vup.getY(), -vup.getZ()
	      );    
    
    // switch in model mode:
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    // reset light sources:
    gpLightCnt = 0;

    for (int i = 0; i < 8; i++) glDisable(GL_LIGHT0 + i);

    // setup the light sources:
    GShadingIterator<GShadeLight> tIter;
    if (gShadeLights) gShadeLights->doWithElements(tIter);
} 

int gGetID(void) {
    static sNumber = 1;
    return(sNumber++);
}

void gFreeID(int) {}

void gShaderUnsetMatrix(const GLookatCamera&, const GPixmapDisplay&) {
    glPopMatrix();
}

void GShadePointLight::shade(void) {}

void GShadeAmbientLight::shade(void) {}

