//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel. Copying, distribution and     
//  legal info is in the file COPYRIGHT which should be distributed with this 
//  file. If COPYRGHT is not available or for more info please contact: 
//		ekki@prakinf.tu-ilmenau.de
/////////// (C) '95 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include "render/grender.h"

GVector3 gCalcVUP(const GVector3& vp, const GVector3& rp, GFloat twist) {
    GVector3 xa, ya, vup;
    GVector3 dr = rp -vp;
    if (dr.mX == 0.0 && dr.mZ == 0.0) {
	xa = GVector3(1,0,0); 
	ya = GVector3(0,0,1); 
    } 
    else { 
	xa = dr.cross(GVector3(0,1,0));
	ya = dr.cross(xa); 
    }
    if (twist == 0.0) vup = ya;
    else {
	GVector3 xan = xa * sqrt(1/(xa.mX * xa.mX + xa.mY * xa.mY + xa.mZ * xa.mZ));
	GVector3 yan = ya * sqrt(1/(ya.mX * ya.mX + ya.mY * ya.mY + ya.mZ * ya.mZ));
	GVector3 d; 
	GFloat length;
	if (twist > 0 && twist <= M_PI_2) {
	    d = xan - yan;
	    length = sin(twist)/sin(M_PI - (M_PI_4 + twist));
	    vup = yan + d * sqrt(length * length/(d.mX * d.mX + d.mY * d.mY + d.mZ * d.mZ));
	}
	if (twist > M_PI_2 && twist <= M_PI) {
	    twist -= M_PI_2;
	    d = yan * (-1)-xan; 
	    length = sin(twist)/sin(M_PI - (M_PI_4 + twist));
	    vup = xan + d * sqrt(length * length/(d.mX * d.mX + d.mY * d.mY + d.mZ * d.mZ));
	}
	if (twist > M_PI && twist <= M_PI + M_PI_2) {
	    twist -= M_PI;
	    d = yan - xan;
	    length = sin(twist)/sin(M_PI - (M_PI_4 + twist));
	    vup = yan * (-1) + d * sqrt(length * length/(d.mX * d.mX + d.mY * d.mY + d.mZ * d.mZ));
	}
	if (twist > M_PI + M_PI_2 && twist <= M_PI + M_PI) {
	    twist -= M_PI + M_PI_2;
	    d = yan - xan * (-1);
	    length = sin(twist)/sin(M_PI - (M_PI_4 + twist));
	    vup = xan*(-1) + d * sqrt(length * length/(d.mX * d.mX + d.mY * d.mY + d.mZ * d.mZ));
	}
    }    
    return(vup);
}

void GLookatCamera::computeMatrix(void) {
    GVector3 dr = (getRefpoint() - getViewpoint()).unitize(); 
    // lookat direction
    GFloat twist = -getTwist() * M_PI / 180.0;
    // the z axis rotation angle 
    
    // compute y vector:
    GVector3 ye;
    ye.setX(0); 
    ye.setY(1); 
    ye.setZ(0);  
    // defaults
    
    // is dr kollinear with ye ?:
    if((dr.getX() == 0) && (dr.getY() == 1) && (dr.getZ() == 0)) { 
	ye.setX(0); 
	ye.setY(0); 
	ye.setZ(1);
    }
    
    if((dr.getX() == 0) && (dr.getY() == -1) && (dr.getZ() == 0)) {
	ye.setX(0); 
	ye.setY(0); 
	ye.setZ(-1);
    }
    
    // compute eye point xe vector:
    GVector3 xe;
    xe = ye.cross(dr); 
    
    // unitize xe vector:
    xe = xe.unitize();
    
    // compute ye vector:
    ye = dr.cross(xe); 
    
    GMatrix4 trRt; // translation and rotation matrix
    trRt.set(0, 0, xe.getX());
    trRt.set(1, 0, xe.getY());
    trRt.set(2, 0, xe.getZ());
    
    trRt.set(0, 1, ye.getX());
    trRt.set(1, 1, ye.getY());
    trRt.set(2, 1, ye.getZ()); 
    
    trRt.set(0, 2, dr.getX());
    trRt.set(1, 2, dr.getY());
    trRt.set(2, 2, dr.getZ());
    
    trRt.set(0, 3, getViewpoint().getX()); 
    trRt.set(1, 3, getViewpoint().getY()); 
    trRt.set(2, 3, getViewpoint().getZ());
    
    trRt.set(3, 0, 0); 
    trRt.set(3, 1, 0); 
    trRt.set(3, 2, 0); 
    trRt.set(3, 3, 1);
    
    GMatrix4 tw;
    tw.set(0, 0, cos(twist));
    tw.set(0, 1, -sin(twist));
    tw.set(1, 0, sin(twist));
    tw.set(1, 1, cos(twist));
    
    mMat = (trRt * tw);
}

GRay3 GLookatCamera::inverseTransform(const GPoint2& pPoint, GFloat pAspect) const {
    if (isPerspectivic()) {
	GVector3 vn;  
	// normalized vector
	vn.setX((-2 * pPoint.getX() + 1) * pAspect); 
	vn.setY(( 2 * pPoint.getY() - 1));
	vn.setZ(1/tan(getZoom()/2.0 * M_PI/180.0));
	
	GVector3 dir = (getMatrix() * vn) - getViewpoint();
	
	GRay3 tRay;
	tRay.setPoint(getViewpoint());
	tRay.setVector(dir.unitize());
	return(tRay);
    }
    else {
	// NOTE: currently twist not supported.
	GFloat tWidth = tan((GFloat)getZoom() * M_PI_2 / 90.0);
	GFloat tHeight = tWidth;
	
	GFloat dx = (2 * pPoint.getX() - 1) * tWidth;
	GFloat dy = (2 * pPoint.getY() - 1) * tHeight;
	if (pAspect > 1) dx *= pAspect;
	else dy /= pAspect;

	GPoint3 tPoint;
	tPoint.setX(getViewpoint().getX() + dx);
	tPoint.setY(getViewpoint().getY() + dy);
	tPoint.setZ(getViewpoint().getZ());
	
	GRay3 tRay;
	tRay.setPoint(tPoint);
	tRay.setVector((getRefpoint() - getViewpoint()).unitize());
	return(tRay);
    }
}
